# Chat

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#    Copyright (C) 2002  Monash University
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#
# note: the cache for channels is buggy, because of multiple threads
# writing the same data
#
#
#


# TODO: pfh
#   remove support for 'chat message' once everyone has upgraded

# Yolande:
#  blue is a bad choice for online people - it looks like a clickable link.
#  the arrows aren't an obvious visual clue to their behaviour.  would checkboxes work better?
#  arrow keys are a bit yucky.
#  gossip

# TODO: (njh)
#   - clickable urls in chat window(James Mollison)
#   - ability to remove people from click-list(Michale Kaufman)

# TODO: detect when someone drops off the network badly

# TODO: float pickle

# TODO: improve /recall : select from list, only recall from some people for single message

# All messages in the chat widget are displayed in "fields",
# eg a field may display a message indicating some task is in progress, 
# then later change to say that it has finished

from __future__ import generators
import cStringIO
import os
import re
import select
import string
import sys
import traceback
import types

import settings
import utility
import error
import chat_commands, safe_pickle, hash
import calendar
import time
import check


# When publishing timestamps, we standardize on the Unix epoch of 1970.
host2standard_time_offset = calendar.timegm((1970, 1, 1, 0, 0, 0, 3, 1, -1))

if time.gmtime(host2standard_time_offset)[:6] != (1970, 1, 1, 0, 0, 0):
    check.show_bug(_('Bug in calculating epoch offset.'))


def host2standard_timestamp(host_tstamp):
    check.check_has_type(host_tstamp, types.FloatType)
    
    ret = long(host_tstamp + host2standard_time_offset)
    
    check.check_has_type(ret, types.LongType)  #=@E16
    return ret

def standard2host_timestamp(std_tstamp):
    check.check_has_type(std_tstamp, types.LongType)  #=@R36
    
    ret = float(std_tstamp - host2standard_time_offset)
    
    check.check_has_type(ret, types.FloatType)  #=@E17
    return ret


def chat_tell_thread(chat,dests,message, field, attachment=None, augmentation=None):
    """Messages are sent in parallel, one thread per recipient. The
       information displayed in field ends with '...' until the last
       message has been sent, then with '..' until all recipients
       have answered, and finally with '.' The field is closed by
       the last thread to finish """
    
    import threading
    check.check_matches(dests, ['text'])
    check.check_is_text(message)

    std_timestamp = host2standard_timestamp(time.time())

    if augmentation and len(augmentation) != len(message):
        sys.stderr.write("Augmentation %d bytes, should be %d bytes.\n" % (len(augmentation), len(message)))
        augmentation = None
    
    try:
        field.show( _('Sending message...\n'), ['grey'])
        result = ['','','','','...\n']
        dest_list = [ ]
        dest_address  = [ ]
        checked_address = [ ]
        dest_names = [ ]
        received_names = [ ]
        offline_list  = [ ]
        offline_names = [ ]
        recipient_list = [ ]
        channel_pipes = { }
        chan_cnt=[0]
        message_lock = threading.RLock()
        
        field_offline=field.child()
        field_offline.show(result[3]+'\n', ['grey'])
        
        for dest in dests:
            if dest[0]=='#':
                channel_pipes[dest] = chat.app.channels.sub_list_pipe(dest)
                recipient_list.append((None,None,dest))
                continue
            acq = chat.app.name_server.locate(dest)
            acq.start_watching(chat.node)
            while not acq.watched:
                yield 'sleep',0.1

            acq.lock.acquire()
            online = acq.online
            address = acq.address
            name = acq.name
            username = acq.info['name']
            acq.lock.release()

            if not online:
                offline_list.append(acq)
                offline_names.append(dest)
                recipient_list.append((name,None,username))
            else:
                dest_list.append(acq)
                dest_address.append(address)
                dest_names.append(dest)
                recipient_list.append((name,address,username))

        my_details = (
            chat.app.name_server.public_key_name,
            chat.app.name_server.name
        )
        #if attachment == None:
        #    sig = chat.app.name_server.sign(message)
        #    message = ('chat message',my_details,recipient_list,message,sig)
        #else:
        #    attachment = safe_pickle.dumps(attachment)
        #    sig = chat.app.name_server.sign(message.encode('utf8')+attachment)
        #    message = ('chat message',my_details,recipient_list,message,attachment,sig)

        package = {
            'from': my_details,
            'to': recipient_list,
            'text': message
        }
        if augmentation:
            package['aug'] = augmentation
        if attachment:
            package['attach'] = attachment

        package_pickle = safe_pickle.dumps(package)
        signature = chat.app.name_server.sign(package_pickle)
        message = ('chat message 2',package_pickle,signature)

        def recipient_thread(chat,address,name,channel,received_names,dest_address,
                             checked_address,chan_cnt,field,result, message=message,
                             message_lock=message_lock):
            succ = 1
            try:
                try:
                    ticket, template, wait = chat.node.call(address,message)
                    if wait: yield ('call',(chat.node,ticket))
                    ret_value = chat.node.get_reply(ticket,template)

                except error.Error:
                    # Recipient has not upgraded yet                    

                    package = safe_pickle.loads(message[1])
                    attachment = package.get('attach',None)
                    my_details = package.get('from')
                    recipient_list = package.get('to')
                    text = package.get('text').encode('latin1', 'replace')

                    if attachment == None:
                        sig = chat.app.name_server.sign(text)
                        message_old = ('chat message',my_details,recipient_list,text,sig)
                    else:
                        attachment = safe_pickle.dumps(attachment)
                        sig = chat.app.name_server.sign(text+attachment)
                        message_old = ('chat message',my_details,recipient_list,text,attachment,sig)

                    ticket, template, wait = chat.node.call(address,message_old)
                    if wait: yield ('call',(chat.node,ticket))
                    ret_value = chat.node.get_reply(ticket,template)

                #Type checking
                if type(ret_value) != types.TupleType:
                    succ = 0
                    if not channel:
                        message_lock.acquire()
                        result[2] = result[2] + name + _(' sent bad reply.\n')
                        message_lock.release()
                        field.show( result[0]+result[4]+result[2], ['grey'])
            except error.Error,err:
                succ = 0
                if not channel:
                    message_lock.acquire()
                    result[2] = result[2] + name + _(' could not be contacted: ')+err.message+'\n'
                    message_lock.release()
                    field.show( result[0]+result[4]+result[2], ['grey'])
                    chat.app.name_server.bad_address(address)
                    
            message_lock.acquire()
            if succ:
                if channel:
                    chan_cnt[0] += 1
                else:
                    received_names.append(name)
                    if ret_value[0]:
                        if ret_value[1] != '':
                            result[2] = result[2] + name + ' ' + ret_value[1] + '\n'
                    else:
                        if ret_value[1] == '':
                            result[2] = result[2] + name + _(' is quiet.\n')
                        else:
                            result[2] = result[2] + name + _(' is quiet: ') + ret_value[1] + '\n'
                            
                if chan_cnt[0] == 0:
                    result[0] = _('Message received by ') + utility.english_list(received_names) 
                elif chan_cnt[0] == 1:
                    result[0] = _('Message received by ')\
                                +utility.english_list(received_names+[ _('1 person')])+result[1]
                else:                            
                    result[0] = _('Message received by ')\
                                +utility.english_list(received_names+[ _('%d people') % chan_cnt[0] ])+result[1]
                field.show( result[0]+result[4]+result[2], ['grey'])

            checked_address.append(address)
            checked_address.sort()
            if result[4]=='..\n' and checked_address == dest_address:
                #the last thread closes the field
                if chan_cnt[0]==0 and received_names == []:
                    result[0] = _('Nobody received your message')
                result[4] = '.\n'
                field.show( result[0]+result[4]+result[2], ['grey'])
                field.close()
            message_lock.release()

        for i in range(len(dest_list)):
            utility.start_thread(recipient_thread(chat,dest_address[i],dest_names[i],0,\
                       received_names,dest_address,checked_address,chan_cnt,field,result))
         
        if channel_pipes:
            if channel_pipes.__len__()>1:
                result[1] = (_(' on channels %s') % utility.english_list(channel_pipes.keys()))
            else:
                result[1] = (_(' on channel %s') % channel_pipes.keys()[0])

        if channel_pipes:
            for chan_name in channel_pipes.keys():
                if not chat.app.channels.cache.has_key(chan_name):
                    chat.app.channels.cache[chan_name]=[]
                else:                    
                    for address in chat.app.channels.cache[chan_name]:
                        if address in dest_address:
                            continue                    
                        dest_address.append(address)
                        utility.start_thread(
                            recipient_thread(chat,address,'',1,received_names,
                                             dest_address,checked_address,chan_cnt,field,result))
                #reset the cache:
                chat.app.channels.cache[chan_name]=[]
                    
                    
        while channel_pipes:
            for chan_name,chan_pipe in channel_pipes.items():
                if chan_pipe.finished():
                    chan_pipe.stop()
                    del channel_pipes[chan_name]
                    continue
                
                for address in chan_pipe.read_all():
                    #update the cache
                    if address not in chat.app.channels.cache[chan_name]:
                        chat.app.channels.cache[chan_name].append(address)
                    if address in dest_address:
                        continue
                    dest_address.append(address)
                    utility.start_thread(recipient_thread(chat,address,'',1,received_names,
                               dest_address,checked_address,chan_cnt,field,result))
                    
            yield 'sleep',0.1

        #we now have launched all the tasks
        message_lock.acquire()
        dest_address.sort()
        result[4]='..\n'
        message_lock.release()
        if checked_address == dest_address:
            if chan_cnt[0]==0 and received_names == []:
                result[0] = _('Nobody received your message')
            result[4] = '.\n'
            field.show( result[0]+result[4]+result[2], ['grey'])
            field.close()

        recall_list = [ ]
        for i in range(len(offline_list)):
            key   = utility.random_bytes(settings.name_bytes)
            lock  = hash.hash_of(key)
            crypt = offline_list[i].encrypt((key, message, std_timestamp))
            data  = {
                'type' : 'offline message',
                'crypt': crypt
            }
            name  = hash.hash_of('offline message '+offline_list[i].name)
            # Loose proof of @R50: @I21, @E22.

            recall_list.append((offline_list[i].nickname,name,data,key))

            publish_result, subthread = chat.app.cache.publish([name],data,lock)
            yield 'wait',subthread
            if publish_result:
                redundancy = publish_result[0]
            else: redundancy = 0
            if redundancy == settings.cache_redundancy:
                str = offline_names[i] + _(' will get your message when next logged on.\n')
            elif redundancy == 0:
                str = _('Could not store message for ') + offline_names[i] + '!\n'
            else:
                str = (offline_names[i]
                       + _(' will probably get your message when next logged on:\n' +\
                           '  Message only stored with redundancy %d.\n') % redundancy)

            result[3] = result[3] + str
            field_offline.show(result[3]+'\n', ['grey'])
            
        field_offline.close()
        chat.recall_list = recall_list
    except error.Error, err:
        field.show(_('Error sending message:\n') + err.message + '\n', ['grey'])
        field.close()


# (If you change these then search for uses and update the
# proofs, or at least add a todo note.)
incoming_message_history_item_tmpl = ('any', types.FloatType)
unread_message_list_item_tmpl = ((types.StringType,),
                                 'opt-address',
                                 types.FloatType)




def chat_receive_thread(chat_obj, request, address, field1, quiet, send_time, add_to_history=1):
    """send_time should be as if from time.time() (i.e. a float expressing seconds
       since the local machine's epoch) rather than the standardized timestamps
       sent over the network (i.e. a long expressing seconds since the Unix epoch
       of 1970)."""

    # TODO: more checking of message format
    # TODO: search for people not in acq list

    check.check_is_opt_address(address)  #=@R33  fixme callers
    check.check_has_type(send_time, types.FloatType)  #=@R35  fixme callers
    # fixme: should probably require that request be a tuple with 5+ elems.
    check.check_has_type(field1, types.InstanceType)

    yield 'sleep',0.01 # just a trick to get into main thread...
    field_middle = None
    field_right = None
    field3 = None
    try:
        if not chat_obj.widget_visible:
            chat_obj.n_unseen_messages = chat_obj.n_unseen_messages + 1
            if chat_obj.n_unseen_messages == 1:
                title = _('1 unseen message')
            else:
                title = _('%d unseen messages') % chat_obj.n_unseen_messages
            chat_obj.app.set_title(title)

        if request[0] == 'chat message 2':
            package = safe_pickle.loads(request[1])
            text = package['text']
            augmentation = package.get('aug')
            has_attachment = package.has_key('attach')
            attachment = package.get('attach')
            sender_key_name, sender_name = package['from']
            recipients = package['to']

            verify_text = request[1]
            verify_sig = request[2]

        else:
            # Legacy message format
            text = request[3]
            augmentation = None
            has_attachment = (len(request) == 6)
            if has_attachment:
                attachment = safe_pickle.loads(request[4])

            recipients = request[2]

            if type(request[1]) == type(()):
                sender_key_name = request[1][0]
                sender_name = request[1][1]
            else:
                # Used not to send username with key name
                sender_key_name = request[1]
                sender_name = 'unknown'

            if has_attachment:
                verify_text = request[3].encode('utf8')+request[4]
                verify_sig = request[5]
            else:
                verify_text = request[3]
                verify_sig = request[4]

        if not augmentation:
            augmentation = chr(128) * len(text)

        field1.show('[', ['grey'])
            
        field_sender = field1.child()

        # Until we work out who they really are...
        field_sender.show( '<'+sender_name+'>', ['grey'])

        field_middle = field_sender.child()
        field1.close()
        field_middle.show( ' >>', ['grey'])

        field_right = field_middle.child()
        field_right.show( ' ...] ', ['grey'])

        field = field_right.child()

        while 1:
            match = re.search('(http:|ftp:|circle-file:)[^ \t\n]*',text)
            if not match:
                break

            match_start, match_end = match.span()

            this_field, field = field, field.child()
            this_field.show(text[:match_start],[],augmentation[:match_start])
            this_field.close()

            this_field, field = field, field.child()
            # lookup circle url here
            if text[match_start:match_start+12] == 'circle-file:':
                try:
                    bad_url=0
                    name=hash.url_to_hash(text[match_start:match_end])
                except:
                    bad_url=1

                if not bad_url:
                    def lookup_url_thread(chat,field,name=name):
                        import search
                        field.set_action(None)
                        pipe = chat.app.node.retrieve(name)
                        root = search.Search_tree_interior(lambda self=chat: pipe, [], [], 'file','',
                                                       '', _(': couldn\'t find anything, sorry'),'')
                        sleep_time=0.1
                        while root.still_searching:
                            yield 'sleep',sleep_time
                            sleep_time=sleep_time*1.25
                            root.update(chat.app.node)
                        if root.get_children():
                            for item in root.get_children():
                                item.set_player(chat.app)
                                field.set_action(chat.file_field_on_click(field,widget,event))
                                field.leaf = item
                                field.show( '%s' % item.get_text()+' ','files')
                                item.field=field.child()
                                item.field.set_action(on_click)
                            field.close()
                        else:
                            field.set_action(chat.url_field_on_click)
                            field.show('%s' % field.text,'links')
                            field.close()
                        pipe.stop()

                    utility.start_thread(lookup_url_thread(chat_obj,this_field,name))
                    this_field.show('%s' % text[match_start:match_end],
                                    [],augmentation[match_start:match_end])
                else:
                    this_field.show(text[match_start:match_end],
                                    [],augmentation[match_start:match_end])
            else:
                this_field.set_action(chat_obj.url_field_on_click)
                this_field.show(text[match_start:match_end],['links'])
                this_field.close()

            text = text[match_end:]
            augmentation = augmentation[match_end:]

        field.show( text, [], augmentation)
        chat_obj.mmll= text
        field3 = field.child()
        field.close()


        now_date  = time.strftime(_('%A %B %d, %Y'),time.localtime(time.time()))
        then_date = time.strftime(_('%A %B %d, %Y'),time.localtime(send_time))

        if now_date != then_date:
            date_str = '\n'+then_date+' '
        else:
            date_str = '  '
        date_str = date_str + time.strftime(_('%I:%M %p'),time.localtime(send_time))
                        
        if has_attachment:
            chat_obj.lock.acquire()
            i = 1
            while chat_obj.exec_vars.has_key("%s%d" % (sender_name, i)):
                i = i + 1
            chat_obj.exec_vars["%s%d"        % (sender_name, i)] = attachment
            chat_obj.exec_vars["%s%d_source" % (sender_name, i)] = address
            chat_obj.lock.release()
            if type(attachment) == type({}) and attachment.get('type',None) == 'file':

                field3_extra = _('\nContains ')
                chat_obj.show_before(field3, date_str + field3_extra, ['light_grey'])
                field3.show('attachment', ['links'])
                field3.set_action(lambda f,w,e: chat_obj.take_field_on_click(
                    f,w,e, ["%s%d" % (sender_name,i)])) # field, widget, event
                field4 = field3.child()
                field4.show(_("; type `!%s%d\' to see its details, type `/take %s%d\' to retrieve\n\n")
                            %(sender_name,i,sender_name,i), ['light_grey'])
                field4.close()
            else:
                field3_extra = _('\nContains attached python object; type `!%s%d\' to see.') % (sender_name,i)
        else:
            field3.show(date_str + '\n\n', ['light_grey'])

        if not quiet and chat_obj.app.config['beep_on_message']:
            utility.beep()

        previous_field = field_middle
        for item in recipients:
            if not check.matches(item, ('any', 'opt-address')):
                sys.stderr.write('chat_receive_task passed bad item in recipients list: '
                                 + `item` + '.  Ignoring this item.\n')
                continue

            field_recipient = previous_field.child()
            previous_field = field_recipient
            recipient_key_name,recip_address = item[0],item[1]
            if recipient_key_name==None and len(item)>=3 and item[2][0]=='#':
                if chat_obj.app.config['auto_list_channels']:  # kc5tja, 22jul2002
                    chat_obj.app.channels.sub(item[2], 1)

                field_recipient.set_action(chat_obj.channel_field_on_click)
                field_recipient.show( ' '+item[2], ['links'])
                field_recipient.close()
                continue

            try:
                result_id, thread = chat_obj.app.name_server.identify(
                    recipient_key_name,recip_address)
                yield 'wait',thread
                recipient = result_id[0]
                # Proof of @R32: recip_address is assigned from item[1] above.
                # item was checked above to be a tuple whose second elem is
                # an opt-address.  Neither item nor recip_address are
                # reassigned.  Deep immutability from @E21.
                
                #chat_obj.refresh_user(recipient.nickname)
                field_recipient.acq=recipient
                field_recipient.set_action(chat_obj.name_field_on_click)
                field_recipient.show(' ' + recipient.nickname, ['people'])
                field_recipient.close()

            except error.Error:
                if len(item) >= 3:
                    bad_nick_str = ' <'+item[2]+'>'
                else:
                    bad_nick_str = ' ?'
                field_recipient.show(bad_nick_str,['grey'])
                field_recipient.close()

            #Avoid excessive updating
            #field1.show(str+' ...]', ['blue'])
        field_right.show('] ', ['grey'])

        try:
            result_id,thread = chat_obj.app.name_server.identify(sender_key_name, address)
            yield 'wait',thread
            sender = result_id[0]
        except error.Error:
            sender = None

        if sender:
            try:
                sender.verify(verify_text,verify_sig)
                field_sender.acq = sender
                field_sender.set_action(chat_obj.name_field_on_click)
                field_sender.show(sender.nickname,'people')
            except error.Error:
                field_sender.show( 'IMPOSTER '+sender.nickname.upper(), 'red')

        field_sender.close()

    except error.Error:
        chat_obj.show(_('Bad message received from ') + repr(address) + '\n')

    for f in (field_middle, field_right, field3):
        if f is not None:
            f.close()

    if add_to_history:
        chat_obj.lock.acquire()
        try:
            new_item = (request, send_time)
            check.check_matches(new_item, incoming_message_history_item_tmpl)
            # Proof: @R35, immutability of float, send_time not reassigned
            # in this function.
            # Relevance: @I16.
            chat_obj.incoming_message_history.append(new_item)

            while len(chat_obj.incoming_message_history) > settings.incoming_message_history_size:
                del chat_obj.incoming_message_history[0]
        finally:
            chat_obj.lock.release()
            

def retrieve_cached_messages_thread(self):
    field = self.get_field()
    field.show(_('Retrieving any messages sent while you were offline...\n'),'grey')

    pipe = self.node.retrieve(
        hash.hash_of('offline message '+self.app.name_server.public_key_name),
        settings.cache_redundancy)

    # Loose proof of @R50: @I22, @E22.
    pipe_reads = [ ]
    if not pipe.finished():
        while 1:
            pipe_reads.extend(pipe.read_all())
            if pipe.finished():
                break
            yield 'sleep',1
    pipe.stop()
    # pjm 2002-08-05: I've changed the above to sleep only if
    # one read_all call isn't enough.  However, I don't know why
    # sleep is wanted in the first place, or whether a different
    # duration might be better.  (Python sleep allows its
    # argument to be fractional, implemented in terms of
    # select.)

    unique_messages = [ ]
    for item in pipe_reads:
        # I think we're guaranteed that item matches ('af_inet_address', 'any').
        if not check.matches(item[1], {'type' : types.StringType,
                                       'crypt' : ('any', 'any')}):
            # print bad peer
            pass
        elif item[1] not in unique_messages:
            unique_messages.append(item[1])

    message_list = [ ]
    for raw_msg in unique_messages:
            if type(raw_msg) == type({}) \
               and raw_msg.get('type','') == 'offline message' \
               and raw_msg.has_key('crypt'):
                try:
                    decrypt = self.app.name_server.decrypt(raw_msg['crypt'])
                    if not check.matches(decrypt, ('text',
                                                   (types.StringType,),
                                                   types.LongType)):
                        raise error.Error('bad decrypted reply')

                    # Remove from caches
                    for thing in pipe_reads:
                        if thing[1] == raw_msg:
                            try:
                                ticket, template, wait = self.node.call(\
                                    thing[0],('data cache remove',decrypt[0]))
                                if wait: yield ('call',(self.node,ticket))
                                self.node.get_reply(ticket,template)

                            except error.Error:
                                pass

                    message_list.append((decrypt[2],decrypt[1]))
                except error.Error:
                    pass

    message_list.sort()
    self.lock.acquire()
    try:
        any = 0
        for msg in message_list:
            if msg not in self.offline_message_buffer:
                self.offline_message_buffer = [msg] + self.offline_message_buffer[:50]
                #utility.start_thread(chat_receive_thread(self,msg[1],None, self.get_field(),1,msg[0]).start()
                new_item = (msg[1], None,
                            standard2host_timestamp(msg[0]))
                # Proof of @R36: msg is taken from message_list.
                # message_list is local to this method, and is not
                # passed to any other method (so is not shared with any
                # other thread).  message_list starts as empty and is
                # written to solely as (decrypt[2],decrypt[1]) pairs,
                # and only where decrypt has already been found to
                # match ('any', ('string',), 'long').  The relevant
                # types are immutable.
                check.check_matches(new_item,
                                    unread_message_list_item_tmpl)
                # Proof: the @R36 proof just above also shows that
                # msg[1] (i.e. decrypt[1]) matches ('string',) and
                # is immutable.  new_item[1] matches because
                # is_opt_address(None).  new_item[2] matches from @E17.
                # Relevance: @R.I15
                self.unread_message_list.append(new_item)
                any = 1
            else:
                print _("Duplicate offline message.")
    finally:
        self.lock.release()

    if any:
        field.show(
            _('\nYou received messages while you were offline. Type /read to read them.\n'),
            'grey')
        self.set_normal_prompt()
    else:
        field.show('\n')
    field.close()


import threading


class Chat:
    """ Base class of Chat_gtk and Chat_text, provides basic chat functions. """

    def __init__(self,app, use_gtk=1):
        #utility.Task_manager.__init__(self)
        self.lock = threading.RLock()
    
        self.node = app.node
        self.app  = app
        self.use_gtk = use_gtk

        #
        self.first_field = None
        self.last_field = None


        # Visible status
        self.quiet       = 0
        self.activity    = ''

        # Detect repeated offline messages
        self.offline_message_buffer = [ ]

        # Store a copy of recent messages
        self.incoming_message_history = [ ]

        # used for /ls
        self.current_path_nickname = ''
        self.current_path = [ ]        

        # While quiet, messages are buffered for later display.
        # Tuples of form (request, opt-address, time.time()).
        # address for offline-sent messages is None.
        # The time item must be wrt this client's epoch.
        self.unread_message_list = [ ]
        
        config = utility.get_checked_config("chat", types.DictionaryType, { })
        for name,tmpl in [('quiet', 'any'),
                          ('activity', 'any'),
                          ('offline_message_buffer', 'any'),
                          ('incoming_message_history', [('any', 'any timestamp')]),
                          ('unread_message_list', [((types.StringType,),
                                                    'opt-address',
                                                    'any timestamp')])]:
            if config.has_key(name):
                config_val = config[name]
                if check.matches(config_val, tmpl):
                    setattr(self, name, config_val)
                else:
                    print (_("Warning: ignoring chat config item %s, which doesn't match %s.")
                           % (name, tmpl))
        # The current version of Circle always writes these timestamps in
        # standardized form, but previous versions wrote a mixture of floats
        # and longs.
        for i in xrange(len(self.incoming_message_history)):
            request,tstamp = self.incoming_message_history[i]
            if type(tstamp) == types.LongType:
                self.incoming_message_history[i] = (request,
                                                    standard2host_timestamp(tstamp))
            check.check_matches(self.incoming_message_history[i],
                                incoming_message_history_item_tmpl)
            # Loose proof: definition of check.is_any_timestamp, @E17.
            # Relevance: @R.I16.
        for i in xrange(len(self.unread_message_list)):
            request,addr,tstamp = self.unread_message_list[i]
            if type(tstamp) == types.LongType:
                self.unread_message_list[i] = (request, addr,
                                               standard2host_timestamp(tstamp))
            check.check_matches(self.unread_message_list[i],
                                unread_message_list_item_tmpl)
            # Loose proof: as above.
            # Relevance: @R.I15.
        
        self.recall_list = [ ]

        # Detect repeated messages
        self.message_buffer = [ ]

        self.channel = [ ]

        self.history = [ '' ] # Command history
        self.history_position = 0

        # Results of /find
        self.identity_list = [ ]
        self.file_list=[ ]

        # Python environment
        self.exec_vars = { 'app' : app, 'node' : app.node }

        # Command registry
        self._commands = { }  # name -> (min n. params, max n. params, function)

        chat_commands.register(self)

        self.chat_check_invar()
    
    def chat_check_invar(self):
        check.check_matches(self.unread_message_list,
                            [unread_message_list_item_tmpl])  #=@I15
        check.check_matches(self.incoming_message_history,
                            [incoming_message_history_item_tmpl])  #=@I16
        check.check_matches(self.channel, ['text'])  #=@I31
    
    def repeated(self, address,call_id):
        if (address,call_id) in self.message_buffer:
            return 1
        self.message_buffer.append((address,call_id))
        if len(self.message_buffer) > 50:
            del self.message_buffer[0]
        return 0

    def start(self):
        #utility.Task_manager.start(self)
        
        self.app.name_server.set_status('chat',(not self.quiet,self.activity))
        
        self.set_normal_prompt()

        self.node.add_handler('chat message 2',self, ('string',  # pickled message details
                                                      ('long',))) # signature
        self.node.add_handler('chat message',self, (('any', # ns.public_key_name
                                                     'any'), # ns.name
                                                    [('any', 'any', 'text')], # recipient list
                                                    'any', # message (probably needs 'text')
                                                    'any')) # attachment or sig
        self.node.add_handler('chat look', self)
        # The expected call/return signatures are () / (types.IntType, 'text')
        # respectively.  But don't bother checking return type, as callers
        # already do their own checking.
        
        self.app.name_server.add_watch_callback(self)

        self.refresh_userlist()

        if self.app.config['show_tips']:
            self.show_tip()

        if self.app.gossip and self.app.config['show_gossip']:
                
            title_field = self.get_field() 
            title_field.show('Fetching gossip...\n', ['grey'])
            gossip_fields = []
            for i in range(5):
                field = self.get_field()
                field.show('\n')
                gossip_fields.append(field)

            def on_gossiped(title_field=title_field,gossip_fields=gossip_fields,self=self,gossip=self.app.gossip):
                gossip.lock.acquire()
                n = 0
                for item in gossip.sorted_wodges()[:100]:
                    if not item[1].collapsed:
                        gossip_fields[n].set_action(self.gossip_field_on_click)
                        gossip_fields[n].show(
                                item[1].wodge['subject'] + '\n',
                                ['links','indented'])
                        n = n + 1
                        if n >= 5: 
                            break
                for i in range(5-n):
                    gossip_fields[4-i].show( '', [])
                    gossip_fields[4-i].close()
                
                gossip.lock.release()

                if n == 0:
                    str = ''
                else:
                    str="Today's hot gossip:\n"
                title_field.show( str, ['bold'])
                title_field.close()

            self.app.gossip.request_update(on_gossiped)

        utility.start_thread(retrieve_cached_messages_thread(self))

        if self.quiet or self.activity != '':
            self.show('\n')
            self.show_status()

    def stop(self):
        #utility.Task_manager.stop(self)
        #self.change_activity('removing callbacks')
        self.node.remove_handler('chat message 2')
        self.node.remove_handler('chat message')
        #self.node.remove_handler('chat look')

        #thomasV: I guess the following should belong to name_server
        #self.app.name_server.remove_watch_callback(self)

        #self.change_activity('preparing config')
        config = { }
        for item in ['quiet','activity','offline_message_buffer']:
            config[item] = getattr(self,item)
        # Convert timestamps to standardized epoch.  The intent is to allow
        # sharing/copying a .circle directory between different operating
        # systems (in particular, between operating systems that have a
        # value of time.gmtime(0), i.e. different epochs).
        config['incoming_message_history'] \
          = map(lambda item: (item[0], host2standard_timestamp(item[1])),
                self.incoming_message_history)
        config['unread_message_list'] \
          = map(lambda item: (item[0], item[1], host2standard_timestamp(item[2])),
                self.unread_message_list)

        #self.change_activity('saving to disk')
        utility.set_config("chat",config)
        #self.change_activity('')

    def handle_watch(self, acq, because_of, **extras):

        acq.lock.acquire()
        try:
            nickname = acq.nickname
            watch    = acq.watch
        finally:
            acq.lock.release()

        if because_of == 'edit':
            self.lock.acquire()
            try:
                new_channel = filter(lambda dest: dest[:1] == '#' or \
                                        self.app.name_server.nicknames.has_key(dest),
                                     self.channel)
            finally:
                self.lock.release()

            self.set_channel(new_channel)
            # Proof of @R49: new_channel subsetof self.channel; @I31.

        elif acq != self.app.name_server.me:
            self.refresh_userlist()
            
        if not watch or (because_of not in [ 'connect','disconnect','status changed'] ):
            return

        self.lock.acquire()
        try:
            if not self.quiet:
                if because_of == 'connect':
                    action = _(' just arrived.\n')
                elif because_of == 'status changed':
                    s = acq.status.get('chat')
                    if not s: return
                    if not s[1]: return
                    action = ' '+s[1]+'\n'
                elif because_of == 'no reply':
                    action = _(' does not reply.\n')
                else:
                    if extras['message']:
                        action = _(' just left: ') + extras['message'] + '\n'
                    else:
                        action = _(' just left.\n')
                self.show(time.strftime(_('%I:%M %p') + ' ',time.localtime(time.time())),'light_grey')
                field = self.get_field()
                field.acq = acq
                field.set_action(self.name_field_on_click)
                field.show(nickname,'people')
                self.show(action,'grey')
        finally:
            self.lock.release()

    def handle(self, request, address,call_id):
        check.check_matches(request, (types.StringType,))  #=@R34
        check.check_is_af_inet_address(address)  #=@R31
        
        self.lock.acquire()
        try:
            if request[0] == 'chat look':
                return (not self.quiet, self.activity)
            if request[0] in ['chat message', 'chat message 2']:
                if request[0] == 'chat message 2':
                    unpacked = safe_pickle.loads(request[1])
                    # TODO check unpacked against template
                    # TODO don't unpack twice (here and in receive task)
                    sender = unpacked['from']
                    recipients = unpacked['to']
                else:
                    sender = request[1]
                    recipients = request[2]
                    
                for recipient in recipients:
                    if recipient[0] == self.app.name_server.public_key_name:
                        break
                    if recipient[2][:1] == '#' and \
                         self.app.channels.is_listening_to(recipient[2]) and \
                         sender[0] != self.app.name_server.public_key_name:
                        break
                else:
                    return None
            
                if not self.repeated(address,call_id):
                    if self.quiet:
                        new_item = (request, address, time.time())
                        check.check_matches(new_item,
                                            unread_message_list_item_tmpl)
                        # Proof: @R34,@R31, and not having modified request or
                        # address, and the relevant types being immutable
                        # (partly by @E11).  time.time() returns float.
                        # Relevance: @R.I15.
                        self.unread_message_list.append(new_item)
                        self.set_normal_prompt()
                        del new_item
                    else:
                        utility.start_thread(chat_receive_thread(
                            self,request,address, self.get_field(),self.quiet,time.time()))
                        # Proof of @R33: @R31; deep immutability by @E11.
                        # Proof of @R35: time.time() returns float.

                return (not self.quiet, self.activity)

        finally:
            self.lock.release()

    def refresh_userlist(self):
        pass

    def set_prompt(self, str):
        raise error.Error("define in subclass")
    
    def show_tip(self):
        import random
        tip_field = self.get_field()
        self.show_before(tip_field,'Tip: ',['bold'])
        tip_field.show(random.choice(settings.tips)+'\n\n', ['links'])
        tip_field.set_action(self.tip_field_on_click)

    def set_normal_prompt(self):
        if self.shell:
            str = ''
        elif not self.use_gtk and self.command_text_lines:
            str = '>'
        else:
            if not self.channel:
                str = '['+self.app.config['name']+']'
            else:
                str = '['+self.app.config['name']+' >> '+string.join(self.channel,' ')+']'
                self.set_prompt(str)

            l = len(self.unread_message_list)
            if l:
                str = str + _(' (%d unread)')%l

        if not self.use_gtk:
            status = utility.get_config('text_status',{ })
            status['unread'] = len(self.unread_message_list)
            utility.set_config('text_status',status)

        self.set_prompt(str)

    def get_field(self, action_dummy=None):
        raise error.Error("define in subclass")

    def get_field_before(self,field,action_dummy=None):
        raise error.Error("define in subclass")

    def get_field_after(self,field,action_dummy=None):
        raise error.Error("define in subclass")
    

    def show(self,str,color='grey',action=None):
        """This function creates and shows a new field at the end of the field list
        """
        field = self.get_field(action)
        field.show(str,color)
        field.close()

    def show_before(self,field,str,color='grey',action=None):
        """This function creates and shows a new field immediately before field.
        """
        previous=self.get_field_before(field,action)
        if previous:
            previous.show(str,color)
            previous.close()

    def show_after(self,field,str,color='grey',action=None):
        """This function creates and shows a new field immediately after field.
        """
        next=self.get_field_after(field,action)
        if next:
            next.show(str,color)
            next.close()


    def show_status(self):
        if self.quiet:
            if self.activity == '':
                self.show(_('You are quiet.\n'))
            else:
                self.show(_('You are quiet: ')+self.activity+'\n')
        else:
            if self.activity == '':
                self.show(_('You are listening.\n'))
            else:
                self.show(_('What you are doing: %s')%self.activity+'\n')

        if self.unread_message_list:
            self.show(_('You have unread messages. Type /read to read them.\n'))

    def remove_channel(self, dest):
        check.check_matches(dest, 'text')  #=@R51
        # Relevance: @R.I31.

        self.lock.acquire()
        try:
            if dest not in self.channel:
                return
            self.channel.remove(dest)
            self.refresh_userlist()
            self.set_normal_prompt()
        finally:
            self.lock.release()

    def set_channel(self, dests):
        check.check_matches(dests, ['text'])  #=@R49
        # Relevance: @R.I31.

        self.lock.acquire()
        try:
            self.channel = dests
            self.app.channels.check_join(dests)
            self.refresh_userlist()
            self.set_normal_prompt()
        finally:
            self.lock.release()

    def set_status(self, quiet,activity):
        self.lock.acquire()
        self.quiet = quiet
        self.activity = activity
        self.lock.release()
        
        self.app.name_server.set_status('chat',(not quiet,activity))

    def send_message(self, dests, message, attachment=None, augmentation=None):
        check.check_matches(dests, ['text'])
        check.check_is_text(message)

        if self.quiet:
            self.show_status()
        utility.start_thread(chat_tell_thread(
            self,dests,message,self.get_field(),attachment,augmentation))

    def register(self, name, min_param, max_param, function):
        """ Register a command. 
        
                A command takes a certain number of parameters separated by spaces,
                up to max_param, and not less than min_param. The last parameter may
                contain spaces.

                function may assert that its argument matches ['text'] (i.e. is
                a list of items each of type string or unicode) and that the
                length of this list is in the range [min_param, max_param].

                E.g.:
                
                def mycommand(chat, params):
                    check.check_matches(params, ['text'])
                    check.check_assertion(1 <= len(params) <= 2)
                    
                    str = 'You typed ' + params[0]
                    if len(params) == 2:
                        str += '; ' + params[1]
                    chat.show(str + '\n')
                app.chat.register('mycommand', 1, 2, mycommand)
                
        """
        check.check_has_type(name, types.StringType)
        check.check_has_type(min_param, types.IntType)
        check.check_has_type(max_param, types.IntType)
        check.check_is_callable(function)
        check.check_assertion(0 <= min_param)
        #check.check_assertion(min_param <= max_param)
        
        self.lock.acquire()
        self._commands[name] = (min_param, max_param, function)
        self.lock.release()
        
    def do(self, str, str_aug=None):
        """ Execute a chat command. """        

        if not str_aug:
            str_aug = chr(128) * len(str)

        #str = string.strip(str)
        while str[:1] in [' ','\n','\t']:
            str = str[1:]
            str_aug = str_aug[1:]
        while str[-1:] in [' ','\n','\t']:
            str = str[:-1]
            str_aug = str_aug[:-1]

        try:
            if not str:
                pass

            elif str[:2] == '//':
                self.run_shell(str[2:])

            elif str[0] == '/':
                # pjm: I'm not sure what we should do if str is just "/" or
                # is "/ blah", and whether the behaviour should depend
                # on whether "/blah" would be a valid command.
                # Usually the user would mean "/say /" (or "/say / blah").
                # Versions of circle up to 0.31 implicitly treat "/ blah"
                # the same as "/blah" (due to the behaviour of string.split).
                if ((str == '/') or (str[1] in string.whitespace)):
                    list = ['', string.lstrip(str[2:])]
                else:
                    list = string.split(str[1:], maxsplit=1)
                command = self._commands.get(list[0])
                if command is None:
                    raise error.Error(_("Unknown command /%s.\n(Did you mean `/say %s...' ?)\n"\
                                        +"Type /help for help on available commands.") % (list[0], str[:8]))

                if len(list) == 2:
                    remainder = list[1]
                else:
                    remainder = ''

                if list[0] in ['ls','play','get']:
                    
                    param = string.split(remainder)
                    # filenames may include whitespaces
                    for i in range(len(param)-1):
                        if param[i][-1:]=='\\':
                            param[i+1] = param[i][:-1]+' '+param[i+1]
                            param[i]=''
                    while '' in param:
                        param.remove('')

                else:
                    if command[1]:
                        param = string.split(remainder,maxsplit=command[1]-1)
                    else:
                        if remainder:
                            raise error.Error(_("/%s doesn't accept parameters.\n"\
                                                +"Type /help for descriptions of commands.") % list[0])
                        param = [ ]

                if len(param) < command[0]:
                    raise error.Error(_('/%s requires at least %d parameters.\n'\
                                        +'Type /help for help on command usage.')%(list[0],command[0]))

                check.check_matches(param, ['text'])
                check.check_assertion(command[0] <= len(param))
                if command[1] != -1:
                    check.check_assertion( len(param) <= command[1])

                if command[2].func_code.co_argcount == 3:
                    # Command uses augmentation
                    apply(command[2],(self,param,str_aug[-len(param[-1]):]))
                else:
                    # Command does not use augmentation
                    apply(command[2],(self,param))
            
            elif str[0] == '!':
                file = cStringIO.StringIO()
                stdout = sys.stdout
                stderr = sys.stderr
                sys.stdout = file
                sys.stderr = file
                try:
                    try:
                        result = eval(str[1:],self.exec_vars)
                        if result is not None:
                            print `result`
                    except SyntaxError:
                        exec str[1:]+'\n' in self.exec_vars
                except:
                    traceback.print_exc()
                    if len(str) > 10:
                        short_str = str[:7] + '...'
                    else:
                        short_str = str
                    print _("(Perhaps you meant `/say %s' ?)" % short_str)
                sys.stdout = stdout
                sys.stderr = stderr
                self.show(file.getvalue())
                
            elif self.channel == [ ]:
                #self.show("Use /join to specify someone to talk to.\n")
                #self.run_shell(str)
                self.show("If you want to send a message, you need to select a person or channel to talk to. Type '/who #welcome' to list the people listening on channel #welcome. Click on their names if you want to talk to them. Type '/join #welcome' to send a message to the whole channel.\n")
            else:
                self.send_message(self.channel[:], str, None, str_aug)
        
        except error.Error, err:
            self.show(err.message+'\n')
        
        self.set_normal_prompt()
    
    def handle_user_entry(self, str, str_aug=None):
        """ Handler user typing some text. """
        if self.shell:
            self.set_prompt('')

            str = str + '\n'
            try:
                while str:
                    n = os.write(self.shell_master,str)
                    str = str[n:]
            except OSError:
                pass
            return

        self.do(str, str_aug)


# vim: expandtab:shiftwidth=4:tabstop=8:softtabstop=4 :
