# gtk chat interface

#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  N Hurst and P Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import generators

if __name__ == '__main__':
    import pygtk
    pygtk.require('2.0')
    
from chat import *
import gtk, pango
import widgets

def _augment(buffer, offset,augmentation):
  iter = buffer.get_iter_at_offset(offset)
  for item in augmentation:
      next_iter = iter.copy()
      next_iter.forward_char()
      buffer.apply_tag_by_name("aug"+str(ord(item)/16), iter, next_iter)
      iter = next_iter


class Chat_field_gtk:
    def __init__(self, chat, start=None, end=None, next=None, action=None):
        """Chat_field_gtk.__init__:
	    chat:  a reference back to the owning chat object
	    start: offset of the start of the field
	    end:   offset of the end of the field
            next:  the next field in the list.

	    action: this function is called when a
            mouse button is clicked in this field.
        """
	check.check_isinstance(chat, Chat)
	if (start is not None or end is not None):
	    check.check_has_type(start, types.IntType)
	    check.check_has_type(end, types.IntType)
        if action is not None:
            check.check_is_callable(action)

        chat.lock.acquire()
        try:
	    self.action = action
	    self.text=''
            self.chat = chat
            self.closed = 0

            if start is not None:
                self.start = start
                self.end   = end
            else:
                self.start = chat.prompt_start
                self.end   = chat.prompt_start

            if next is not None:
                self.next = next
                self.prev = next.prev
                next.prev = self
                if self.prev:
                    self.prev.next = self
            else:
                self.next = None
                self.prev = chat.last_field
                if chat.last_field is not None:
                    chat.last_field.next=self

                if chat.first_field is None:
                    chat.first_field = self
                chat.last_field=self
                
        finally:
            chat.lock.release()
    

    def set_action(self,action=None):
        if action is not None:
            check.check_is_callable(action)
        self.action = action


    def child(self,action=None):
        """ Only function that can be called from other thread (except __init__)
         njh: Sure, but wtf does it do?
         pfh: Creates a field located immediately after this field
         """
        if action is not None:
            check.check_is_callable(action)

        if self.closed:
            return
        
        self.chat.lock.acquire()
        try:
            return Chat_field_gtk(self.chat, self.end, self.end, self.next, action=action)
        finally:
            self.chat.lock.release()


    def offset_successors(self,change):
        # Update these now so apply-tag event is not confused
        self.chat.prompt_end = self.chat.prompt_end + change
        self.chat.prompt_start = self.chat.prompt_start + change

        #offset all fields after me
        item = self.next
        while item is not None:
            item.start = item.start + change
            item.end   = item.end   + change
            item = item.next
        

    def show_gtk(self,str,tags=[],augmentation=None):
        """display text in the field. to be called from gtk thread only"""
        check.check_is_gtkthread()
        
        if self.closed:
            return
        if type(tags) != type([]):
            tags = [tags]
        if tags == []:
            tags = [ "black" ]
        if type(str) == type(''):
            str = unicode(str,'ascii','replace')

        self.text = str
        change = len(str) - (self.end - self.start)
        self.offset_successors(change)

        self.gtk_show(str,tags,augmentation,self.start, self.end)
        self.end = self.end + change


    def show(self,str,tags=[],augmentation=None):
        """display text in the field. to be called from non gtk threads"""
        
        if self.closed:
            return
        if type(tags) != type([]):
            tags = [tags]

        if tags == []:
            tags = [ "black" ]

        if type(str) == type(''):
            str = unicode(str,'ascii','replace')

        self.text = str
        change = len(str) - (self.end - self.start)
        self.offset_successors(change)

        gtk.idle_add(self.gtk_show,str,tags,augmentation,self.start, self.end)
        self.end = self.end + change

    def apply_tags(self,tags):
        check.check_is_gtkthread()
        for tag in tags + ['clickable']:
            self.chat.buffer.apply_tag_by_name(
                tag,
                self.chat.buffer.get_iter_at_offset(self.start),
                self.chat.buffer.get_iter_at_offset(self.end))
       

    def gtk_show(self,str,tags,augmentation, start, end):
        """Internal function: never use it, always use show """
        check.check_is_gtkthread()
        self.chat.lock.acquire()
        cursor_was_visible = self.chat.is_cursor_visible()
        if start != end:
            self.chat.buffer.delete(
                self.chat.buffer.get_iter_at_offset(start),
                self.chat.buffer.get_iter_at_offset(end))
        # insert_with_tags_by_name broken, so do it ourselves
        self.chat.buffer.insert(
            self.chat.buffer.get_iter_at_offset(start),
            str)
        self.chat.buffer.remove_all_tags(
            self.chat.buffer.get_iter_at_offset(start),
            self.chat.buffer.get_iter_at_offset(start+len(str)))
        for tag in tags + ['clickable']:
            self.chat.buffer.apply_tag_by_name(
                tag,
                self.chat.buffer.get_iter_at_offset(start),
                self.chat.buffer.get_iter_at_offset(start+len(str)))
        if augmentation:
            _augment(self.chat.buffer, start, augmentation)

        #apply(self.chat.buffer.insert_with_tags_by_name,
        #    [ self.chat.buffer.get_iter_at_offset(self.start),
            #  str ] + tags)
        if cursor_was_visible:
            self.chat.force_cursor_onscreen()
        self.chat.lock.release()


    def close(self):
        """ close field. """

        #do not close clickable fields...
        if self.action is not None:
            return

        self.closed=1
	self.chat.lock.acquire()

        if self==self.chat.first_field:
            self.chat.first_field=self.next
        if self==self.chat.last_field:
            self.chat.last_field=self.prev
        if self.prev:
            self.prev.next= self.next
        if self.next:
            self.next.prev = self.prev
                
        self.chat.lock.release()


class Chat_gtk(Chat):
    """ GTK chat interface. """

    def get_insertion_offset(self):
        insertion_mark = self.buffer.get_insert()
        return self.buffer.get_iter_at_mark(insertion_mark).get_offset()

    def is_cursor_visible(self):
        insertion_mark = self.buffer.get_insert()
        y, height = self.view.get_line_yrange(
            self.buffer.get_iter_at_mark(insertion_mark) )
        visible = self.view.get_visible_rect()

        return y <= visible.y+visible.height and y+height >= visible.y

    def force_cursor_onscreen(self):
	self.view.scroll_to_mark(self.buffer.get_insert(),0,gtk.FALSE,0,0)

    def on_begin_user_action(self, buffer):
        self.buffer_user_active = 1

    def on_end_user_action(self, buffer):
        self.buffer_user_active = 0

    def on_key_press(self, _w,_e):
        self.key_presses[_e.keyval] = (_e.time, self.get_insertion_offset() - self.prompt_end)

        if _e.string and not _e.string.isalpha():
            self._reaugment_word(self.get_insertion_offset())

        if _e.keyval == 65293: # Enter

            if self.mode == 'edit_mode':
                self.mode = 'prompt_mode'
                text = widgets.get_uslice(
                    self.buffer.get_iter_at_offset(self.edited_field.start),
                    self.buffer.get_iter_at_offset(self.edited_field.end))
                self.file_field_on_edit(self.edited_field,text)
                self.view.emit_stop_by_name("key-press-event")
                return
                
            if self.completion_field:
                self.completion_field.show_gtk('')
                self.completion_field = None
            offset = self.get_insertion_offset()
            if offset < self.prompt_end:
                return

            if widgets.get_uslice(
                 self.buffer.get_iter_at_offset(offset-1),
                 self.buffer.get_iter_at_offset(offset)
               ) == '\\':
                return

            command = widgets.get_uslice(self.buffer.get_iter_at_offset(self.prompt_end),
                                         self.buffer.get_end_iter())
            command_aug = self.augmentation
            if command:
                self.history[-1] = command
                self.history.append('')
                self.history_position = len(self.history)-1

	    self.prompt_start = self.buffer.get_end_iter().get_offset()
	    self.prompt_end = self.prompt_start
            self.augmentation = ''
            self.show('\n') # note: gtk_show was needed here
	    self.buffer.place_cursor(self.buffer.get_end_iter())
            
            #command = string.replace(command,'\\\n','\n')
            min_index = 0
            while 1:
                index = command.find('\\\n')
                if index < min_index: break
                min_index = index
                command = command[:index] + command[index+1:]
                command_aug = command_aug[:index] + command_aug[index+1:]
            
            self.view.emit_stop_by_name("key-press-event")
            self.handle_user_entry(command, command_aug)
            return
 
        if _e.state & 5: # Ctrl

            if self.shell != None:
           
                if _e.keyval == ord('c'): #Ctrl-C
                    self.lock.acquire()
                    try:
                        os.kill(self.shell,2)
                        self.view.emit_stop_by_name("key-press-event")
                    except OSError:
                        pass
                    self.lock.release()
                return

            prefix = widgets.get_uslice(self.buffer.get_iter_at_offset(self.prompt_end),
                                        self.buffer.get_end_iter())


            # this pushes prefix in the stack
            # useful if you want to postpone a message or command
            if _e.keyval == ord('o'): #Ctrl-O
                if prefix:
                    self.history[-1] = prefix
                    self.history.append('')
                    self.history_position = len(self.history)-1
                    self.buffer.delete(self.buffer.get_iter_at_offset(self.prompt_end),
                                       self.buffer.get_end_iter())

            if prefix:
                return
            
            if _e.keyval == ord('x'): #Ctrl-X
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/exit ')
                self.lock.release()

            elif _e.keyval == ord('d'): #Ctrl-D
                self.run_shell('exec $SHELL')
                self.show('\n')

            elif _e.keyval == ord('s'): # Ctrl-S            
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/search ')
                self.lock.release()

            elif _e.keyval == ord('f'): # Ctrl-F            
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/find ')
                self.lock.release()
            
            elif _e.keyval == ord('j'): # Ctrl-J            
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/join ')
                self.lock.release()                        

            elif _e.keyval == ord('m'): # Ctrl-M
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/me ')
                self.lock.release()

            elif _e.keyval == ord('w'): # Ctrl-M
                self.lock.acquire()
                self.buffer.insert(self.buffer.get_end_iter(),'/who ')
                self.lock.release()

            elif _e.keyval == ord('l'): # Ctrl-L
                self.clear()
                self.set_normal_prompt()

            #elif _e.keyval == ord('a') and (_e.state & 5): # Ctrl-A
            #    self.buffer.place_cursor(cmd_line)
            #    return gtk.TRUE
            #
            #elif _e.keyval == ord('k') and (_e.state & 5): # Ctrl-K
            #    eol = insertion.copy()
            #    eol.forward_line()
            #    self.buffer.delete(insertion, eol)
            #    return gtk.TRUE


        if _e.keyval == 65289: #tabulation: auto-completions
            
            prefix = widgets.get_uslice(self.buffer.get_iter_at_offset(self.prompt_end),
                                        self.buffer.get_end_iter())
            if self.shell != None:
                # todo: reading the terminal's auto-completion does not work yet
                self.lock.acquire()
                self.view.emit_stop_by_name("key-press-event")
                os.write(self.shell_master,prefix+'\t')
                self.prompt_start = self.buffer.get_end_iter().get_offset()
                self.prompt_end = self.prompt_start
                self.buffer.place_cursor(self.buffer.get_end_iter())
                self.lock.release()
            else:
                # auto-completion in the chat window: for nicknames, commands and file names.
                if not prefix:
                    if self.completion_field:
                        self.completion_field.show('')
                    return gtk.TRUE
                if prefix[0] !='/':
                    if self.completion_field:
                        self.completion_field.show('')
                    return gtk.TRUE
                
                words=prefix.split()
                # filenames may include whitespaces
                for i in range(len(words)-1):
                    if words[i][-1:]=='\\':
                        words[i+1] = words[i][:-1]+' '+words[i+1]
                        words[i]=''
                while '' in words:
                    words.remove('')


                choices = []                
                last_word=words[-1:][0]
                if prefix[-1:]==' ':
                    if prefix[-2:] !='\\ ':
                        last_word=''
                
                if len(words) == 1 and prefix[-1:] != ' ':
                    choices=self._commands.keys()
                    for i in range(len(choices)):
                        choices[i]='/'+choices[i]
                    suffix=' '
                else:                
                    command = prefix.split()[0]
                    if command not in ['/ls','/cd','/join','/forget','/remember','/look','/listen','/get','/play','/help','/who']:
                        return gtk.TRUE

                    if command in ['/ls','/cd','/join','/forget','/remember','/look','/listen']:
                        choices=self.app.name_server.nicknames.keys()
                        for id in self.identity_list:
                            nick=id.item['name']
                            if nick not in choices:
                                choices.append(nick)

                    if command in ['/join','/listen','/who']:
                        for ch in self.app.channels.list.keys():
                            if ch not in choices:
                                choices.append(ch)

                    if command in ['/help']:
                        choices=self._commands.keys()
                        for i in range(len(choices)):
                            choices[i]='/'+choices[i]
                            suffix=' '
                                                    
                    for i in range(len(choices)):
                        if command in ['/ls','/cd']:
                            choices[i]=choices[i]+':'
                            suffix=''
                        else:
                            choices[i]=choices[i]
                            suffix = ' '

                    if command in ['/ls','/cd']:
                        choices.append(':')
                        import search
                        for dir in search.visited_directories:
                            if dir.find(last_word)==0:
                                z = dir[len(last_word):]
                                colon_index = z.find(':')
                                slash_index = z.find('/')
                                if colon_index !=-1:
                                    z = z[:colon_index+1]
                                elif slash_index != -1:
                                    z= z[:slash_index+1]
                                completion = last_word+z
                                if completion not in choices:
                                    choices.append(completion)
                        
                    if command in ['/play','/get']:
                        for file in self.file_list:
                            filename = utility.force_unicode(file.item['filename'])
                            if filename not in choices:
                                choices.append(filename)
                        suffix=' '

                choices.sort()
                root, completions = utility.complete(last_word,choices)
                if len(completions)>1:
                    str=''
                    for i in completions:
                        colon_index = last_word.rfind(':')
                        slash_index = last_word.rfind('/')
                        if slash_index !=-1:
                            i = i[slash_index+1:]
                        elif colon_index != -1:
                            i = i[colon_index+1:]
                        str=str+i+'   '
                    if self.completion_field == None:
                        self.completion_field=self.get_field()
                    self.completion_field.show(str+'\n','grey')
                else:
                    if self.completion_field:
                        self.completion_field.show('')
                    
                if last_word == '':
                    suffix = ''
                    
                if len(completions)!=1:
                    suffix = ''

                self.lock.acquire()
                i=self.buffer.get_end_iter()
                self.buffer.insert(i,root[len(last_word):].replace(' ','\\ ')+suffix)
                self.lock.release()
                
            return gtk.TRUE


        position = self.get_insertion_offset() - self.prompt_end
        if position >= 0 and _e.keyval in [65362, 65364]:
            command = widgets.get_uslice(self.buffer.get_iter_at_offset(self.prompt_end),
                                        self.buffer.get_end_iter())
	    
            if string.find(command[:position],'\n') == -1 and \
               _e.keyval == 65362: # Up arrow
                if self.history_position > 0:
                    self.history[self.history_position] = command
                    self.history_position = self.history_position - 1
		    str = self.history[self.history_position]
                    self.buffer.delete(
                        self.buffer.get_iter_at_offset(self.prompt_end),
                        self.buffer.get_end_iter())
                    self.buffer.insert(self.buffer.get_end_iter(), str)
		    self.force_cursor_onscreen()
                return 1

            if string.find(command[position:],'\n') == -1 and \
                 _e.keyval == 65364: # Down arrow
                if self.history_position < len(self.history)-1:
                    self.history[self.history_position] = command
                    self.history_position = self.history_position + 1
                    str = self.history[self.history_position]
                    self.buffer.delete(
                        self.buffer.get_iter_at_offset(self.prompt_end),
                        self.buffer.get_end_iter())
                    self.buffer.insert(self.buffer.get_end_iter(), str)
                    self.buffer.place_cursor(self.buffer.get_iter_at_offset(self.prompt_end))
		    self.force_cursor_onscreen()
                return 1

    def on_key_release(self, _w,_e):
        if not self.app.config['augmented_text']:
            return

        if not self.key_presses.has_key(_e.keyval):
            return

        press_time, offset = self.key_presses[_e.keyval]
        del self.key_presses[_e.keyval]

        string = utility.force_unicode(_e.string)
        if not string or \
           ord(string) < 32:
             return

        if offset < 0 or offset >= len(self.augmentation):
            return

        if string.isalpha():
            interval = (_e.time - press_time) / 1000.0
            level = int( interval*4 *255)
            if level < 0: level = 0
            if level > 255: level = 255
        else:
            level = 128

        tag = "aug" + str(level/16)
        iter1 = self.buffer.get_iter_at_offset(self.prompt_end+offset)
        iter2 = iter1.copy()
        iter2.forward_char()
        self.augmentation = self.augmentation[:offset] + \
                            chr(level) + \
                            self.augmentation[offset+1:]
        self.buffer.remove_all_tags(iter1,iter2)
        self.buffer.apply_tag_by_name(tag,iter1,iter2)

    def on_scroll(self, widget, event):
        adj = self.widget.get_vadjustment()
        if event.direction == gtk.gdk.SCROLL_UP:
            adj.set_value(adj.value - adj.step_increment)
        if event.direction == gtk.gdk.SCROLL_DOWN:
            adj.set_value(adj.value + adj.step_increment)
        
    def on_mark_set(self, buffer, iter, mark):
        if mark.get_name() == 'insert' and \
           self.prompt_start <= self.buffer.get_iter_at_mark(mark).get_offset() < self.prompt_end:
            self.buffer.move_mark(mark, self.buffer.get_iter_at_offset(self.prompt_end))

    def on_click2(self, widget, event):
        print "event",dir(event),event.type,event.get_coords()

    def on_click(self, tag, widget, event, iter):

        if event.type != gtk.gdk.BUTTON_RELEASE and event.type != gtk.gdk.BUTTON_PRESS :            
            return gtk.FALSE

        # Is there a selection? If yes, we won't deal with it.
        insert = self.buffer.get_iter_at_mark(
            self.buffer.get_insert()).get_offset()
        selection_bound = self.buffer.get_iter_at_mark(
            self.buffer.get_selection_bound()).get_offset()
        if insert != selection_bound:
            return gtk.FALSE

        # test if we clicked in a field
        field= self.last_field
        while field:
            if field.start <= iter.get_offset() < field.end:
                if field.action != None:
                    result = field.action(field,widget,event)
                    return gtk.TRUE                
                else:
                    return gtk.FALSE                
            field=field.prev

        # Is there a dragged item?
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 2 and self.drag:
            self.paste_file_field(self.drag)
            self.drag = None
            self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
            return gtk.TRUE
            
        return gtk.FALSE
    
    def on_delete(self, _w,start,end):
        if self.buffer_user_active and self.mode == 'prompt_mode' and \
	   start.get_offset() < self.prompt_end:
            self.buffer.emit_stop_by_name("delete-range")
            return

        if self.mode == 'edit_mode':
            if start.get_offset() < self.edited_field.start:
                self.buffer.emit_stop_by_name("delete-range")
                return
            if end.get_offset() > self.edited_field.end:
                self.buffer.emit_stop_by_name("delete-range")
                return
            change = - end.get_offset() + start.get_offset()
            self.edited_field.offset_successors(change)
            self.edited_field.end += change
            
        if self.mode == 'prompt_mode':
            start_offset = start.get_offset() - self.prompt_end
            end_offset = end.get_offset() - self.prompt_end
            if start_offset < 0:
                return
            self.augmentation = self.augmentation[:start_offset] + \
                                self.augmentation[end_offset:]


    def _on_insert_helper(self, offset, str, length):
        offset = offset - self.prompt_end
        if offset < 0:
            return

        self.augmentation = self.augmentation[:offset] + \
                            chr(128)*len(str) + \
                            self.augmentation[offset:]

    def _reaugment_word(self, word_end_offset):
        if not self.app.config['augmented_text']:
            return

        word_start_offset = word_end_offset

        # TODO: do purely with iters?
        while widgets.get_uslice(
                  self.buffer.get_iter_at_offset(word_start_offset-1),
                  self.buffer.get_iter_at_offset(word_start_offset)
              ).isalpha():
            word_start_offset -= 1

        if word_start_offset == word_end_offset:
            return

        iter1 = self.buffer.get_iter_at_offset(word_start_offset) 
        iter2 = self.buffer.get_iter_at_offset(word_end_offset) 

        word_start_offset -= self.prompt_end
        word_end_offset -= self.prompt_end
        length = word_end_offset - word_start_offset

        sum = 0
        for char in self.augmentation[word_start_offset:word_end_offset]:
            sum += ord(char)

        average = sum / length
        tag = "aug" + str(average/16)

        self.augmentation = self.augmentation[:word_start_offset] + \
                            chr(average) * length + \
                            self.augmentation[word_end_offset:]
        
        self.buffer.remove_all_tags(iter1,iter2)
        self.buffer.apply_tag_by_name(tag,iter1,iter2)

    def on_insert(self, _w, iter, str, length):
        str = utility.force_unicode(str)

        if not self.buffer_user_active:
            self._on_insert_helper(iter.get_offset(),str,length)
	    return

        if self.mode == 'prompt_mode':
            if iter.get_offset() < self.prompt_end:
                self.buffer.emit_stop_by_name("insert-text")
                self.buffer.place_cursor(self.buffer.get_iter_at_offset(self.prompt_end))
                self.buffer.insert(self.buffer.get_iter_at_offset(self.prompt_end), str)
                return
            self._on_insert_helper(iter.get_offset(),str,length)

        if self.mode == 'edit_mode':
            if iter.get_offset() < self.edited_field.start:
                self.buffer.emit_stop_by_name("insert-text")
                self.buffer.place_cursor(self.buffer.get_iter_at_offset(self.edited_field.start))
                return
            if iter.get_offset() > self.edited_field.end:
                self.buffer.emit_stop_by_name("insert-text")
                self.buffer.place_cursor(self.buffer.get_iter_at_offset(self.edited_field.start))
                return

            change = len(str)
            self.edited_field.offset_successors(change)
            self.edited_field.end += change

    def on_draw(self, widget, event):
        self.widget_visible = (event.state != gtk.gdk.VISIBILITY_FULLY_OBSCURED)
        if self.widget_visible:
            self.n_unseen_messages = 0
            self.app.set_title()

    def on_apply_tag(self, widget, tag, iter1, iter2):
        if tag.get_property('name')[:3] != 'aug' and \
           iter1.get_offset() >= self.prompt_end:
            self.buffer.emit_stop_by_name("apply-tag")
    
    def create_tags (self):
        def insert_one_tag_into_buffer(buffer, name, *params):
            tag = gtk.TextTag(name)
            while(params):
                tag.set_property(params[0], params[1])
                params = params[2:]
            table = buffer.get_tag_table()
            table.add(tag)

        
        insert_one_tag_into_buffer(self.buffer,
                                   "italic",
                                   "style", pango.STYLE_ITALIC)
        insert_one_tag_into_buffer(self.buffer,
                                   "bold",
                                   "weight", pango.WEIGHT_BOLD)  
        insert_one_tag_into_buffer(self.buffer,
                                   "indented",
                                   "left-margin", 25)  
        insert_one_tag_into_buffer(self.buffer,
                                   "monospace",
                                   "font-desc", pango.FontDescription('monospace'))  

        insert_one_tag_into_buffer(self.buffer,
                                   "softly",
                                   "foreground", "gray30")
        insert_one_tag_into_buffer(self.buffer,
                                   "grey",
                                   "foreground", "gray40")  
        insert_one_tag_into_buffer(self.buffer,
                                   "light_grey",
                                   "foreground", "grey50")  
        insert_one_tag_into_buffer(self.buffer,
                                   "quiet",
                                   "foreground", "chartreuse4")  
        insert_one_tag_into_buffer(self.buffer,
                                   "people",
                                   "foreground", "blue")  
        insert_one_tag_into_buffer(self.buffer,
                                   "files",
                                   "foreground", "blue")  
        insert_one_tag_into_buffer(self.buffer,
                                   "links",
                                   "foreground", "blue")  
        insert_one_tag_into_buffer(self.buffer,
                                   "black",
                                   "foreground", "black")  
        insert_one_tag_into_buffer(self.buffer,
                                   "active",
                                   "foreground", "purple")  
        insert_one_tag_into_buffer(self.buffer,
                                   "red",
                                   "foreground", "red")  
        insert_one_tag_into_buffer(self.buffer,
                                   "error",
                                   "foreground", "purple")  
        insert_one_tag_into_buffer(self.buffer,
                                   "hilite",
                                   "background", "DarkSeaGreen1")  
        insert_one_tag_into_buffer(self.buffer,
                                   "not_editable",
                                   "editable", gtk.FALSE)


        for i in range(16):
            r = max(0, 255*(i-8)/7)
            g = max(0, 196*(8-i)/8)
            b = max(0, 196*(8-i)/8)
            insert_one_tag_into_buffer(self.buffer,
                                       "aug"+str(i),
                                       "foreground", "#%02x%02x%02x"%(r,g,b))  
 
        tag = gtk.TextTag("clickable")

        # gtk may not like empty tags (?)
        tag.set_property("editable", gtk.TRUE)

        tag.connect("event", self.on_click)
        table = self.buffer.get_tag_table()
        table.add(tag)


    def reset_tabs(self):
        self.tabs.set_tab(0,pango.TAB_LEFT,50)
        self.tabs.set_tab(1,pango.TAB_LEFT,60)
        self.tabs.set_tab(2,pango.TAB_LEFT,80)
        self.tabs.set_tab(3,pango.TAB_LEFT,70)
        self.view.set_tabs(self.tabs)
        self.maxlen=2


    def __init__(self,app):
        Chat.__init__(self,app,1)
        
        self.lock.acquire()
        self.completion_field = None
        
	self.prompt_end = 0
	self.prompt_start = 0
        self.drag = None       # drag and drop object
        
	self.buffer_user_active = 0 # Are we processing events from a user action?
        self.augmentation = '' # Key press intencity for chars in user entry area
        self.key_presses = { } # When and where were keys pressed (use in key release to update augmentation)

        self.sw = gtk.ScrolledWindow(None, None)
        self.sw.set_policy(gtk.POLICY_AUTOMATIC,
                           gtk.POLICY_NEVER)
	self.sw.set_size_request(100,100)
        #self.sw.add_mask(gtk.gdk.BUTTON_PRESS_EVENT)
        
        self.view = gtk.TextView()
        self.view.set_left_margin(5)
        self.view.set_right_margin(5)
        self.view.set_wrap_mode(gtk.WRAP_WORD)

        self.tabs=pango.TabArray(4,1)
        self.reset_tabs()

        self.mode = 'prompt_mode'

        #self.colors = {}
        
        self.buffer = self.view.get_buffer()
        self.create_tags() # build up the tag table for this buffer
        
        self.sw.add(self.view)
        self.widget = self.sw
        
        self.userlist_widget = gtk.ScrolledWindow()
        self.userlist_widget.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)

        self.userlist_viewport = gtk.Viewport()
        self.userlist_widget.add(self.userlist_viewport)

        self.view.connect("key-press-event",self.on_key_press)
        self.view.connect("key-release-event",self.on_key_release)
        self.view.connect("scroll-event",self.on_scroll)

        #this is for clicking in empty regions:
        #self.view.connect("event", self.on_click)

	self.buffer.connect("begin-user-action",self.on_begin_user_action)
	self.buffer.connect("end-user-action",self.on_end_user_action)
        self.buffer.connect("delete-range",self.on_delete)
        self.buffer.connect("insert-text",self.on_insert)
        self.buffer.connect("mark-set",self.on_mark_set)
        
        #self.buffer.connect("apply-tag",self.on_apply_tag)
        
        self.widget_visible = 1
        self.n_unseen_messages = 0
        
        self.view.connect("visibility-notify-event",self.on_draw)
        self.view.set_events(gtk.gdk.VISIBILITY_NOTIFY_MASK)

        self.apply_colors()

        # Shell command support
        self.shell = None   # (pid, write_pipe)
        
        try:
            openpty = None
            try:
                from pty import openpty
            except:
                from os import openpty
            self.shell_master, self.shell_slave = openpty()
            
            def callback(source, condition, fd=self.shell_master, self=self):
                """
                there is one chat field for each running command
                escape commands are allowed within that field
                """
                try:
                    str = os.read(fd,256)
                except OSError:
                    str = ''
                #str = string.replace(str,'\r','')
                
                #for the moment we do not have local tabs
                #therefore we better convert tabs into spaces
                str = string.replace(str,'\t','    ')
                
                if string.find(str,chr(7)) != -1:
                    str = string.replace(str,chr(7),'')
                    utility.beep()

                #here I can insert xml tags, (for colors, for example)
                
                new_text = self.shell_field.text
                #print "ww: ",str
                m = re.search(r'\x1B\[\d*;?\d*[a-zA-Z@:#~ ]', str)
                while m:
                    escape = m.group()
                    print "escape",escape[1:]
                    if escape[:-1] == 'H':
                        print "home"
                        if escape == '\x1B[H':
                            index = 0
                        elif re.match(r'\x1B\[\d*;\d*H'):
                            index = 0

                    new_text = new_text+str[0:m.start()]
                    str = str[m.end():]
                    m = re.search(r'\x1B\[\d*;?\d*[a-zA-Z@:#~ ]', str)
                
                #print "ww2: ",str
                new_text = new_text + str
                utility.force_unicode(new_text)
                    
                self.shell_field.show(new_text, ['monospace'])
                #note: if I detect a prompt I should not display it and I should allocate a new field
                return 1

            self.shell_tag = gtk.input_add(self.shell_master,gtk.gdk.INPUT_READ,callback)
        
            import termios

            attr = termios.tcgetattr(self.shell_slave)
            attr[3] = attr[3] & ~termios.ECHO
            termios.tcsetattr(self.shell_slave,termios.TCSANOW,attr)
        except:
            self.shell_master = None
            self.shell_slave  = None
            
        self.lock.release()

    def handle(self, request, address, call_id):
        check.check_matches(request, (types.StringType,))  #=@R40
        check.check_is_af_inet_address(address)

        return Chat.handle(self, request, address, call_id)


    def show(self,str,color='grey',action=None):
        """
        This function creates and shows a new field
        at the end of the field list
        it must be called in the gtk thread
        """
        field = self.get_field(action)
        field.show(str,color)
        field.close()

    def apply_colors(self):

        self.lock.acquire()
        
        (r, g, b) = self.app.config['background_color']
        col = self.view.get_colormap().alloc_color(r,g,b)
        self.view.modify_base(gtk.STATE_NORMAL,col)

        (r, g, b) = self.app.config['text_color']
        col = self.view.get_colormap().alloc_color(r,g,b)
        self.view.modify_text(gtk.STATE_NORMAL,col)

        self.lock.release()

        def apply_color(self,color_name):
            (r,g,b) = self.app.config[color_name+'_color']
            tag = self.buffer.get_tag_table().lookup(color_name)
            tag.set_property("foreground", "#%02x%02x%02x"%(r>>8,g>>8,b>>8))
            
        apply_color(self,"people")
        apply_color(self,"files")
        apply_color(self,"links")
        apply_color(self,"quiet")
        apply_color(self,"active")

        (r_text,g_text,b_text) = self.app.config['text_color']
        (r_bg  ,g_bg  ,b_bg  ) = self.app.config['background_color']
        
        r = (int(r_text *0.6 + r_bg *0.4))
        g = (int(g_text *0.6 + g_bg *0.4)) 
        b = (int(b_text *0.6 + b_bg *0.4)) 
        tag = self.buffer.get_tag_table().lookup("grey")
        tag.set_property("foreground", "#%02x%02x%02x"%(r>>8,g>>8,b>>8))

        r = (int(r_text *0.3 + r_bg *0.7)) 
        g = (int(g_text *0.3 + g_bg *0.7))
        b = (int(b_text *0.3 + b_bg *0.7)) 
        tag = self.buffer.get_tag_table().lookup("light_grey")
        tag.set_property("foreground", "#%02x%02x%02x"%(r>>8,g>>8,b>>8))

        for i in range(16):
            r = max(0, (r_text >>8) + 255*(i-8)/7)
            r = min(255,r)
            g = max(0, (g_text >>8) + 196*(8-i)/8)
            g = min(255,g)
            b = max(0, (b_text >>8) + 196*(8-i)/8)
            b = min(255,b)
            tag= self.buffer.get_tag_table().lookup("aug"+str(i))
            tag.set_property("foreground", "#%02x%02x%02x"%(r,g,b))


    def refresh_userlist(self):
        """Uses gtk."""
        def refresh_userlist_task(self):
            import gtk
            import widgets

            widget = gtk.VBox(0,0)

            channels = self.app.channels.list.keys()
            channels.sort()
            for name in channels:
                try:
                    ch = self.app.channels.list[name]
                except ValueError:
                    continue
                
                hbox = gtk.HBox(0,0)

                mute = widgets.Toggle_icon('<')
                mute.set_relief(gtk.RELIEF_NONE)
                def on_toggle(mute, name=name,self=self):
                    self.lock.acquire()
                    self.app.channels.mute(name,not mute.get_active())
                    self.view.grab_focus()
                    self.lock.release()
                mute.connect("toggled", on_toggle)
                self.lock.acquire()
                mute.set_active(not ch['muted'])
                self.lock.release()
                hbox.pack_start(mute,0,0,0)
                widgets.tooltips.set_tip(mute,_("listen to channel %s")%name,"")
                
                checker = widgets.Toggle_icon('>')
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                            self.app.channels.mute(name,0)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    self.view.grab_focus()
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk on %s")%name,"")

                n = name
                if len(n) > 36:
                    n = name[0:33]+"..."
                
                button = widgets.Text_button(n) 
                button.modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("blue"))# set_style(self.blue_style)
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.edit_settings(self.app.channels,name)
                def on_press(button, ev, name=name,self=self):
                    if ev.button == 3:
                        channel_context_menu(self,self.app.channels,name, ev)
                button.connect("clicked",on_click)
                button.connect("button-press-event",on_press)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s settings")%name)

                widget.pack_start(hbox,0,0,0)
 
            self.app.name_server.lock.acquire()
            try:
                list = [ ]
                for item in self.app.name_server.nicknames.items():
                    if item[1] is not self.app.name_server.me:
                        if item[1].watch:
                            dist = item[1].distance
                            if dist == None: dist = 100.0
                        else:
                            dist = 101.0
                        if item[1].online:
                            top_sorter = 0
                        else:
                            top_sorter = 1
                        list.append(((top_sorter,dist),item[0],item[1]))
                list.sort()
            finally:
                self.app.name_server.lock.release()

            for item in list:
                item[2].lock.acquire()
                try:
                    name    = item[2].nickname
                    online  = item[2].online
                    status  = item[2].status.get('chat',(1,''))
                finally:
                    item[2].lock.release()

                addendum = ''
                if status[0] and status[1]:
                    addendum = '\n('+name+' '+status[1]+')'
                elif not status[0] and status[1]:
                    addendum = '\n('+name+_(' is quiet: ')+status[1]+')'
                
                hbox = gtk.HBox(0,0)
 
                checker = widgets.Toggle_icon('>', 3)
                checker.set_relief(gtk.RELIEF_NONE)
                def on_toggle(checker, name=name,self=self):
                    self.lock.acquire()
                    if checker.get_active():
                        if name not in self.channel:
                            self.channel.append(name)
                    else:
                        if name in self.channel:
                            self.channel.remove(name)
                    self.view.grab_focus()
                    self.set_normal_prompt()
                    self.lock.release()
                checker.connect("toggled", on_toggle)
                self.lock.acquire()
                checker.set_active(name in self.channel)
                self.lock.release()
                hbox.pack_start(checker,0,0,0)
                widgets.tooltips.set_tip(checker,_("talk to ")+name+addendum,"")
                    
                button = widgets.Text_button(name) 
                button.set_relief(gtk.RELIEF_NONE)
                def on_click(button, name=name,self=self):
                    self.app.name_server.lock.acquire()
                    try:
                        if self.app.name_server.nicknames.has_key(name):
                            self.app.edit_acquaintance(
                                self.app.name_server.nicknames[name],self.app.name_server)
                    finally:
                        self.app.name_server.lock.release()

                def on_press(button, ev, self=self, name=name):
                    name_context_menu(button,ev,self,name)
                    
                button.connect("clicked",on_click)
                button.connect("button-press-event",on_press)
                hbox.pack_start(button,1,1,0)
                widgets.tooltips.set_tip(button,_("view %s's details")%name+addendum)
                
                if online:
                    if status[0]:
                        #button.set_style(self.blue_style)
                        button.modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("blue"))
                    else:
                        #button.set_style(self.green_style)
                        button.modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("chartreuse4"))
                    
                widget.pack_start(hbox,0,0,0)
             
            widget.show_all()
            for child in self.userlist_viewport.get_children():
                self.userlist_viewport.remove(child)
            self.userlist_viewport.add(widget)

        gtk.idle_add(refresh_userlist_task, self)


    def get_field(self, action=None):
        """Creates and returns a new field located at the end of the list"""
        if action is not None:
            check.check_is_callable(action)
        return Chat_field_gtk(self,action=action)

    def get_field_from_text(self,begin,end,action=None):
        """field in the command line. use with caution"""
        return Chat_field_gtk(self, begin, end, action=action)

    def get_field_before(self,field,action=None):
        """Creates and returns a new field located before field"""
        if action is not None:
            check.check_is_callable(action)
        if field.closed:
            return
        return Chat_field_gtk(self, field.start, field.start, field, action=action)


    def get_field_after(self,field,action=None,url=''):
        """Creates and returns a new field located after field"""
        if action is not None:
            check.check_is_callable(action)
        if field.closed:
            return
        return Chat_field_gtk(self, field.next.start, field.next.start, field.next, action=action)

    

    def clear(self):
        self.lock.acquire()
        #self.view.set_iter(self.view.get_length())
        self.buffer.delete(self.buffer.get_start_iter(),
                           self.buffer.get_end_iter())
        
        item=self.first_field
        while item is not None:
            item.closed=1
            item=item.next
        self.first_field = None
        self.last_field = None
        self.completion_field = None
        self.prompt_start = 0
        self.prompt_end = 0

        self.file_list = []
        self.file_list_nickname=''
        self.file_list_path=''
        
        self.lock.release()
        self.show("\n")
        self.reset_tabs()

    def delete_fields(self,field_start,field_end):
        """delete an interval of fields"""
        self.lock.acquire()
        start_iter=self.buffer.get_iter_at_offset(field_start.start)
        end_iter=self.buffer.get_iter_at_offset(field_end.end)
        change=field_end.end-field_start.start
        self.buffer.delete(start_iter,end_iter)

        prev = field_start.prev
        next = field_end.next
        if prev:
            prev.next=next
        else:
            self.first_field=next
        if next:
            next.prev=prev
        else:
            self.last_field=prev
            
        item=next
        while item:
            item.start = item.start - change
            item.end   = item.end   - change
            item=item.next

        item=field_start
        field_end.next=None
        while item:
            item.closed=1
            item=item.next
            
        self.prompt_start = self.prompt_start-change
        self.prompt_end = self.prompt_end-change
        self.lock.release()

    def edit_field(self,leaf):
        # place the cursor in the field,
        # do not let it out of the area until validation
        self.edited_field = leaf.file_field
        self.edited_field.apply_tags(["active"])
        self.mode = 'edit_mode'
        self.buffer.place_cursor(self.buffer.get_iter_at_offset(self.edited_field.start))


    def set_prompt(self,str):

        tags = ['people']
        if self.quiet:
            tags = ['quiet']

        prompt_start = self.prompt_start
        field = Chat_field_gtk(self, self.prompt_start, self.prompt_end)

        #field = self.get_field()
        #self.fields.append(field)
        # Proof of @R45: as per @L1.
        # Proof of @R47: self isinstance Chat_gtk; @I26,@I27.
        # Proof of @R46: index unspecified, its default is None.
        
        field.show(str+' ', tags) #!start, end might change
        #field.start = field.end
        #field.gtk_show(' ', ['aug8'])
        field.close()
        
        self.lock.acquire()
        self.prompt_start = prompt_start
        self.prompt_end = field.end
        self.lock.release()


    def set_prompt_gtk(self, str):
        """runs in the gtk thread : should not create a field..."""
        # Kind of hacked...
        # check.check_is_gtkthread()
        
        #self.set_prompt(str)
        utility.mainthread_call(self.set_promptuiui,str)

       
    def run_shell(self, command):

        if self.shell_master == None:
            self.show( _("No openpty in pty module.\n"))
            return

        self.shell_field = self.get_field()

        try:
            pid = os.fork()
        except:
            self.show(_("Could not fork.\n"))
            return        

        if not pid:
            os.dup2(self.shell_slave, 0)
            os.dup2(self.shell_slave, 1)
            os.dup2(self.shell_slave, 2)
            os.execlp('/bin/bash','/bin/bash','-c',command)
            os._exit(0)

        #S = self.sync()
        self.shell = pid
        #del S

        def timeout(self=self,pid=pid):
            result = os.waitpid(pid,os.WNOHANG)
            if result[0] == pid:
                #S = self.sync()
                self.shell = None
                #del S

                self.set_normal_prompt()
                return 0
            return 1

        gtk.timeout_add(250, timeout)


    def set_tabs(self,length):
        """The first column is the title or filename (variable width)
           The second is the size for files (width 70)
           The third is the mimetype (width 90)
        """
        import pango
        if length>self.maxlen:
            self.maxlen=length
            if self.maxlen>90:
                self.maxlen=90
            self.tabs.set_tab(0,pango.TAB_LEFT,self.maxlen*7)
            self.tabs.set_tab(1,pango.TAB_LEFT,self.maxlen*7+70)
            self.tabs.set_tab(2,pango.TAB_LEFT,self.maxlen*7+70+90)
            self.view.set_tabs(self.tabs)


    def paste_file_field(self,leaf):
        #str = ' '+leaf.item['filename']+' '
        str = ' circle-file:'+hash.hash_to_url(leaf.item['name'])+' '

        begin = self.get_insertion_offset()
        if self.prompt_end > begin:
            begin=self.prompt_end
        end = begin + len(str)

        buffer = self.buffer
        iter = buffer.get_iter_at_offset(begin)
        buffer.insert_with_tags_by_name(iter,str,"files")
        
        field = self.get_field_from_text(begin,end,self.file_field_on_click)
        field.leaf = leaf
        field.show(str,"files")



    def name_field_on_click(self,field,widget,event):

        def make_acq_thread(self,field):
            try:
                result, thread = self.app.name_server.make_acquaintance(field.data)
                yield 'wait',thread
                field.acq = result[0]
            except error.Error, err:
                self.app.show_error(err,)
                return

        if not field.acq:
            utility.start_thread(make_acq_thread(self,field))
            #while loop here..        
            #nickname = acq.syncly(lambda acq=acq: acq.nickname)
            nickname = field.data['name']
        else:
            nickname = field.acq.nickname
        
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            if nickname not in self.channel:
                self.channel.append(nickname)
            else:
                self.channel.remove(nickname)
            self.set_normal_prompt()
            self.refresh_userlist()
            return gtk.TRUE
        elif event.type == gtk.gdk.BUTTON_PRESS and event.button==3:
            name_context_menu(widget,event,self,nickname)
            return gtk.TRUE


    def channel_field_on_click(self,field,widget,event):
        app=self.app
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            name=field.text.strip()
            self.lock.acquire()
            if name not in self.channel:
                self.channel.append(name)
            else:
                self.channel.remove(name)
            if app.channels.list.has_key(name):
                self.refresh_userlist()
            self.set_normal_prompt()
            #self.widget.set_position(self.widget.get_length())
            self.lock.release()
        elif event.type == gtk.gdk.BUTTON_RELEASE and event.button == 3:
            channel_context_menu(self,app.channels,field.text.strip(), event)

    def url_field_on_click(self,field,widget,event):
        app=self.app
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            try:
                if field.text[0:12]=='circle-file:':
                    import search
                    search.search_for_name(hash.url_to_hash(field.text),'Files matching url', app.node,app)
                else:
                    utility.browse_url(field.text)
            except error.Error, err:
                self.app.show_error(err,)
            field.show(field.text,['links'])
            return gtk.TRUE
        else:
            return gtk.FALSE

    def host_field_on_click(self,field,widget,event):
        if event.type == gtk.gdk.BUTTON_RELEASE :            
            import search
            str=utility.force_string(field.text.split()[0])
            address=str.split(':')
            search.search_browse_files((address[0],int(address[1])),self.app.node,self.app)
            return gtk.TRUE
        else:
            return gtk.FALSE

    def file_field_on_edit(self,field,text):
        #todo:file_server should do that
        import os
        old_path = field.leaf.item.get('local_path')
        if not old_path:
            return
        new_path = os.path.join(os.path.split(old_path)[0],text)
        os.rename(old_path,new_path)
        field.leaf.item['local_path'] = new_path
        field.show(text,'files')

    def file_field_on_click(self,field,widget,event):

        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 2:
            self.drag = field.leaf
            self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
            return gtk.TRUE
        
        elif event.type == gtk.gdk.BUTTON_RELEASE and event.button == 2:

            if self.drag:
                file_field = self.drag.file_field
                file_field.show(file_field.text,['grey'])
                file_field.set_action(None)
                #file_field.show(_('moved.\n'),['grey'])
                self.app.file_server.move_file(self.drag,field.leaf.father)                

                #if field.leaf.father.expanded:
                #    field.leaf.father.show_expanded(self,field.leaf.father.field.child())
                
                self.drag = None
                self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
                return gtk.TRUE

        elif event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            try:
                playername = field.leaf.player.players.items()[0][0]
            except:
                playername=''
            if playername:
                field.leaf.play(playername,self.app)
                #field.leaf.field.show("playing\n",['grey'])
                #field.show(field.text,['active'])
            else:
                if not field.leaf.item.get('local_path'):
                    if not self.app.file_server.paths.has_key(field.leaf.item['name']):
                        downloader = self.app.file_server.download(
                            field.leaf,self.app.daemon.config['download_dir'])
            return gtk.TRUE
        elif event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            file_context_menu(widget,event,self.app,field)
            return gtk.TRUE
        else:
            return gtk.FALSE


    def directory_field_on_click(self,field,widget,event,root):

        if event.type == gtk.gdk.BUTTON_PRESS and event.button == 2:
            self.drag = field.leaf
            self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
            return gtk.TRUE

        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 2:
            #if self.drag:
            #    self.app.file_server.move_file(self.drag,root)                
            #    file_field = self.drag.file_field
            #    file_field.show(file_field.text,['grey'])
            #    file_field.set_action(None)
            #    self.drag.field.show(_('moved.\n'),['grey'])
            #    self.drag = None
            #    self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
            #    return gtk.TRUE

            if self.drag:
                file_field = self.drag.file_field
                file_field.show(file_field.text,['grey'])
                file_field.set_action(None)
                file_field.show(_('moved.\n'),['grey'])
                self.app.file_server.move_file(self.drag,root)                

                #if root.expanded:
                #    root.show_expanded(self,root.field.child())
                
                self.drag = None
                self.view.get_window(gtk.TEXT_WINDOW_TEXT).set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
                return gtk.TRUE

        elif event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            if root.expanded:
                self.delete_fields(root.first_field,root.last_field)
                root.first_field=None
                root.last_field=None
                root.expanded=0
            else:
                root.expanded=1
                root.show_expanded(self,root.field.child())
            return gtk.TRUE
        elif event.type == gtk.gdk.BUTTON_PRESS and event.button == 3:
            directory_context_menu(widget,event,self.app,root,field)
            return gtk.TRUE
        else:
            return gtk.FALSE

    def gossip_field_on_click(self,field,widget,event):
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            #Really should scroll to relevant wodge, until then
            #field.show(field.text,['purple','indented'])
            self.app.show_gossip()

    def tip_field_on_click(self,field,widget,event):
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            import random
            field.show(random.choice(settings.tips)+'\n\n', ['links'])

    def take_field_on_click(self,field,widget,event, what):
        if event.type == gtk.gdk.BUTTON_RELEASE and event.button == 1:
            chat_commands.do_take(self, what )


    def edit_settings(self, channels, channel):

        channels.lock.acquire()
        try:
            window = gtk.Window()
            window.set_border_width(10)
            window.set_title(_('Channel: ')+channel)

            vbox = gtk.VBox(0,5)
            window.add(vbox)

            label = gtk.Label(channel)
            vbox.add(label)

            #label = gtk.Label('(Here should go the subscriber list)')
            label = gtk.Label('')
            vbox.add(label)

            hbox = gtk.HBox(0,5)

            button = gtk.Button(_("Remove from sidebar"))
            def unsub(_b,_ch=channel,_s=channels,_w=window):
                _s.unsub(_ch)
                _w.destroy()
            button.connect("clicked",unsub)
            hbox.pack_start(button, 1,1,0)

            button = gtk.Button(_("List subscribers"))
            def list(_b,_ch=channel,_s=channels):
                import search
                search.search_for_people_on_channel(_ch, _s.node,_s.app)
            button.connect("clicked",list)
            hbox.pack_start(button, 1,1,0)

            button = gtk.Button(_("Cancel"))
            button.connect("clicked",lambda _b,_window=window:_window.destroy())
            hbox.pack_start(button, 1,1,0)

            vbox.pack_start(hbox, 0,0,0)
            
            channels.app.show_window(window, _('channel'))

        finally:
            channels.lock.release()



def channel_context_menu(self, channels, channel, event):
    menu = []
    if channels.list.has_key(channel):

        if channel in self.channel:
            def untalk(data, channels=channels, channel=channel):
                if channel in self.channel:
                    self.remove_channel(channel)

            menu.append((_("Stop talking on %s")%channel, untalk))
        else:
            if channels.list[channel]['muted']:
                def unmute(data, channels=channels, channel=channel):
                    channels.mute(channel, 0)
                menu.append((_("Listen to %s")%channel, unmute))
            else:
                def mute(data, channels=channels, channel=channel):
                    channels.mute(channel, 1)
                menu.append((_("Mute %s")%channel, mute))

            def talk(data, channels=channels, channel=channel):
                if not channel in channels.app.chat.channel:
                    channels.mute(channel, 0)
                    self.channel.append(channel)
                    self.refresh_userlist()
                    self.set_normal_prompt()
            menu.append((_("Talk on %s")%channel, talk))

        if self.channel!=[] and self.channel!=[channel]:
            def talkonly(data, channels=channels, channel=channel):
                channels.mute(channel, 0)
                self.set_channel([channel])
            menu.append((_("Talk only on %s")%channel, talkonly))

        def unsub(data, _s=channels, _ch=channel):
            _s.unsub(_ch)
        menu.append((_("Remove %s from sidebar")%channel,unsub))

    else:

        def sub(data, _s=channels, _ch=channel):
            _s.sub(_ch)
        menu.append((_("Subscribe to %s")%channel,sub))

    def list(_b,_ch=channel,_s=channels):
        import search
        search.search_for_people_on_channel(_ch, _s.node,_s.app)
    menu.append((_("List %s subscribers")%channel, list))

    menuwidget = gtk.Menu()
    for (text, action) in menu:
        mi = gtk.MenuItem(text)
        mi.connect("activate", action)
        mi.show()
        menuwidget.append(mi)
    menuwidget.popup(None,None,None,event.button,event.time)


def name_context_menu(widget,ev,chat,name):
    """name should be an acquaintance nickname"""
    
    if ev.button == 3:
        menu = []

        if name in chat.channel:
            def untalk(data, chat=chat, name=name):
                if name in chat.channel:
                    chat.channel.remove(name)
                    chat.refresh_userlist()
                    chat.set_normal_prompt()
            menu.append((_("Stop talking to %s")%name,untalk))
        else:
            def talk(data, chat=chat, name=name):
                if name not in chat.channel:
                    chat.channel.append(name)
                    chat.refresh_userlist()
                    chat.set_normal_prompt()
            menu.append((_("Talk to %s")%name,talk))
        if chat.channel!=[] and chat.channel!=[name]:
            def talkonly(data, chat=chat, name=name):
                chat.channel = [name]
                chat.refresh_userlist()
                chat.set_normal_prompt()
            menu.append((_("Talk only to %s")%name,talkonly))
        def details(data, chat=chat, name=name):
            chat.app.name_server.lock.acquire()
            try:
                chat.app.edit_acquaintance(
                    chat.app.name_server.nicknames[name],
                    chat.app.name_server)
            finally:
                chat.app.name_server.lock.release()
        menu.append((_("View %s's details")%name, details))
        if chat.app.name_server.nicknames[name].online:
            def browse(data, chat=chat, name=name):
                import search
                search.search_browse_files(
                    chat.app.name_server.nicknames[name].address,
                    chat.app.name_server.node,
                    chat.app)
            menu.append((_("Browse %s's files")%name, browse))
        if not chat.app.name_server.nicknames[name].remember:
            def remember(data, chat=chat, name=name):
                acq = chat.app.name_server.nicknames[name]
                acq.lock.acquire()
                acq.remember = 1
                acq.lock.release()
            menu.append((_("Remember %s")%name, remember))
        def forget(data, chat=chat, name=name):
            chat.app.name_server.forget(name)

        menu.append((_("Forget %s")%name, forget))
        menuwidget = gtk.Menu()
        for (text, action) in menu:
            mi = gtk.MenuItem(text)
            mi.connect("activate", action)
            mi.show()
            menuwidget.append(mi)
        menuwidget.popup(None,None,None,ev.button,ev.time)


def file_context_menu(widget,ev,app,field):
    leaf = field.leaf
    menu = []        
    try:
        playername=leaf.player.players.items()[0][0]
    except:
        playername=''
    # Todo: should create one menu item per player

    if playername:
        if app.music_manager.field == field:
            def stop_playing(data,leaf=leaf,playername=playername):
                app.music_manager.stop_music()
            menu.append((_('Stop playing'), stop_playing))
        else:
            def play(data,leaf=leaf,playername=playername):
                #field.show(field.text,"active")
                #leaf.field.show("playing\n",['grey'])
                leaf.play(playername,app)
            menu.append((_("%s")%playername, play))

            if leaf.player.players.has_key('Append'):
                def play_later(data,leaf=leaf):
                    leaf.play('Append',app)
                menu.append((_("Append to playlist"), play_later))

    #check if the file is local:
    if not leaf.item.get('local_path'):
        for d in app.file_server.downloaders:
            if d.data['name']==leaf.item['name']:
                def cancel(data,d=d):
                    d.cancel_download()
                    d.stop()
                menu.append((_("Cancel download"), cancel))
                break
        else:
            if not app.file_server.paths.has_key(leaf.item['name']):
                def download(data,leaf=leaf):
                    downloader=app.file_server.download(leaf, app.daemon.config['download_dir'])
                menu.append((_("Download"), download))

    if leaf.item.get('name'):
        def paste_url(data,leaf=leaf,app=app):
            app.chat.paste_file_field(leaf)        
        menu.append((_("Paste url"), paste_url))

    if not leaf.item.get('local_path'):
        def view_sources(data,leaf=leaf):
            leaf.view_sources(app)
        menu.append((_("View sources"), view_sources))

    if leaf.item.get('local_path'):
        def delete(data,leaf=leaf):
            os.unlink(leaf.item['local_path'])
            field.show(field.text,['grey'])
            leaf.field.show(_('deleted.\n'),['grey'])
            field.set_action(None)
        menu.append((_("Delete"), delete))
        def rename(data,leaf=leaf):
            app.chat.edit_field(leaf)
        menu.append((_("Rename"), rename))
   
    elif app.file_server.paths.has_key(leaf.item['name']):
        def delete(data,leaf=leaf):
            #note: in order to delete the line I guess I need to have all text in fields
            app.file_server.remove_file(leaf.item['name'])
            field.show(field.text,['grey'])
            leaf.field.show(_('deleted.\n'),['grey'])
            leaf.field.close()
        menu.append((_("Delete local copy"), delete))

        #def move_to(data,leaf=leaf):
        #menu.append((_("move to"), move_to))

    menuwidget = gtk.Menu()
    for (text, action) in menu:
        mi = gtk.MenuItem(text)
        mi.connect("activate", action)
        mi.show()
        menuwidget.append(mi)
    menuwidget.popup(None,None,None,ev.button,ev.time)



def directory_context_menu(widget,ev,app,leaf,field):
    menu = []        

    def hide(data,root=leaf,chat=app.chat,field=field):
        chat.delete_fields(root.first_field,root.last_field)
        root.first_field=None
        root.last_field=None
        root.expanded=0
        field.show(field.text,['files'])
    def show(data,root=leaf,chat=app.chat,field=field):
        root.expanded=1
        root.show_expanded(chat,root.field.child())
    def download(data,root=leaf):
        root.download(app)
    def play(data,root=leaf):
        root.download(app,play=1)
        
    if leaf.expanded:
        menu.append((_("Minimize"), hide))
    else:
        menu.append((_("Expand"), show))
        
    if not leaf.item.get('local_path'):
        menu.append((_("Download directory"), download))
    menu.append((_("Play directory"), play))

    if leaf.item.get('local_path'):
        def delete(data,leaf=leaf):
            try:
                os.rmdir(leaf.item['local_path'])
                field.show(field.text,['grey'])
                field.set_action(None)
            except OSError,err:
                print "cannot delete this directory",err
                pass
        menu.append((_("Delete"), delete))
        def rename(data,leaf=leaf):
            app.chat.edit_field(leaf)
        menu.append((_("Rename"), rename))


    menuwidget = gtk.Menu()
    for (text, action) in menu:
        mi = gtk.MenuItem(text)
        mi.connect("activate", action)
        mi.show()
        menuwidget.append(mi)
    menuwidget.popup(None,None,None,ev.button,ev.time)


def generate_random_messages(chat):
    print "mo"
    f = chat.get_field()
    f.show("Moo")
    f.show("Moo moo")
    f.show(u"foo \u2312 bar")
    f.show("")
    return 1

if __name__ == '__main__':
    import gtk
    window = gtk.Window()
    window.set_default_size(400,400)
    class dud:
        node = 0
	config = { 'name': 'test' }
        def set_title(self):
            pass
    
    app = dud()
    app.none = None
    chat = Chat_gtk(app)
    chat.set_prompt("[hello]")
    chat.show("This is a test.")
    chat.show(u"foo \u2312 bar")
    window.add(chat.widget)
    window.connect("destroy", gtk.mainquit)
    window.show_all()
    gtk.timeout_add(2030, generate_random_messages, chat)
    gtk.mainloop()




# vim: expandtab:shiftwidth=4:tabstop=8:softtabstop=4 :
