
#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
    Searching for various things in the hashtable.    
    GUI stuff: This is used by the current GTK and text interfaces
"""

# TODO
#   "availabe from n sources including njh"
#   "trust person x's taste music"

from __future__ import generators
import sys,os,string,types

import error
import file_server
import name_server
import settings
import utility
import hash

if settings.gtk_loaded():
    import widgets
    import gtk

visited_directories=[]


class Player:
    """
    each filetype has a number of viewers
    a config file can describe the correspondence
    (or look into gnome)
    """
    
    def __init__(self, filename,app):

        # these import lines should really be hidden in class Player
        if not settings.gtk_loaded():
            print "bug: gtk should be loaded past this point"
            
        import mp3
        import textviewer
        try:
            import imageviewer
        except:
            imageviewer = None
    
        self.players = {}

        if sys.platform == 'win32':
            return
    
        extension = string.split(filename, ".")[-1]
        lext = string.lower(extension)

        if lext in ['mp3','ogg']:
            self.players['Play'] = app.play_music_now
            self.players['Append'] = app.play_music_later

        elif lext in ['jpg','jpeg','gif','png','xpm','bmp']:
            if imageviewer:
                self.players['View'] = imageviewer.imageviewer

        elif lext == 'txt':
            self.players['View'] = textviewer.Text_Viewer

        elif lext in ['html', 'htm']:
            self.players['View'] = textviewer.External_Viewer

        elif lext in ['pdf']:
            self.players['View'] = textviewer.External_Viewer        
      
        #self.playable = len(self.players.items()) != 0

    def actions(self):
        return self.players.keys()
        
    def play(self, action, info, downloader,field=None):
        return self.players[action](info, downloader, field)
        
        #if len(self.players.keys()) == 1:
        #  return self.players.values()[0](f, s)
        #else: # display a list of players
        #  pass

def search_browse_files(address, node,app):
    pipe = node.search_browse_task(address, [])
    Searcher('Files',pipe, node,app).start()

def search_for_files(str, node, app):
    temp_list = string.split(str)
    list = [ ]
    for item in temp_list:
        if len(item) >= settings.min_search_keyword_len:
            list.append(utility.remove_accents(item.lower()))

    if not list:
        raise error.Error(_("Please specify some keywords to search for.\n" +\
                            "(keywords must contain at least %d letters)") \
                          % settings.min_search_keyword_len)

    largest,keywords,anti_keywords,title = utility.parse_keywords(list)
    if not keywords:
        raise error.Error(_("Please specify at least one positive search term (i.e. not preceded by !).\n" ))
    pipe = node.retrieve(hash.hash_of(largest))
    #if app.overnet:
    #    app.overnet.retrieve(largest,pipe)
    Searcher(_('Files matching: ')+title,pipe, node,app,keywords,anti_keywords).start()

def search_for_name(name,title, node,app):
    pipe = node.retrieve(name)
    Searcher(title,pipe,node,app).start()

def search_for_people(str, node,app):
    list = string.split(utility.remove_accents(str.lower()))
    if not list:
        raise error.Error(_("Please specify some keywords to search for.\n" +\
                            "For example, their name or username."))
    largest,keywords,anti_keywords,title = utility.parse_keywords(list)
    pipe = node.retrieve(hash.hash_of('identity-name '+largest), settings.identity_redundancy)
    Searcher(_('People matching: ')+title,pipe, node,app,keywords,anti_keywords).start()

def search_for_all_people(node,app):
    pipe = node.retrieve(hash.hash_of('service identity'), settings.identity_redundancy)
    Searcher(_('All people'),pipe, node,app).start()

def search_for_all_channels(node,app):
    pipe = node.retrieve(hash.hash_of('channel exists'), settings.channel_redundancy)
    Searcher(_('All channels'),pipe, node,app).start()

def search_for_people_on_channel(channel, node,app):
    pipe = node.retrieve(hash.hash_of('channel subscribe '+channel), settings.channel_redundancy)
    Searcher(_('People on %s')%channel,pipe, node,app).start()


def search_show_sources(source_list, node,app, title='Sources'):
    pipe = node.search_address_list_task(source_list)
    Searcher(title,pipe, node, app).start()


# Search tree classes

# All such classes must provide "get_children", "get_text", "get_description", "get_actions", "get_widget", "get_comparator", "update" and "stop"

class Search_tree_object:
    def __init__(self):
        pass

    def stop(self):
        pass

    def get_children(self):
        return None

    def update(self, app):
        return 0

    def get_description(self, node_address):
        return ''

    def get_actions(self,app):
        return [ ]

    def get_widget(self):
        return None

    def show(self, depth, chat,field):
        pass

    def get_comparator(self):
        return (-1e10, 0, string.lower(self.text)) 

    def matches(self, item,address):
        return 0


class Search_tree_interior(Search_tree_object):
    def __init__(self, pipe_getter, keywords, anti_keywords, what, mime,\
                 text='',empty_text=' (None)',single_text=' (1 element)',\
                 online_only=0, show_url=0, show_number=0, remote_only=0):
        
        Search_tree_object.__init__(self)
        self.sources = None #thomas
        
        self.keywords = keywords
        self.remote_only = remote_only
        self.nickname=''
        self.anti_keywords = anti_keywords
        self.mime=mime
        self.what=what
        self.text=text
        self.field = None
        self.status_field = None         # to display changing information, like download status
        self.empty_text=empty_text
        self.single_text=single_text
        self.online_only=online_only
        self.show_number=show_number
        self.show_url=show_url
        
        self.pipe_getter = pipe_getter
        self.pipe = None
        
        self.children = [ ]
        self.still_searching = 1
        self.list = [ ] #the children that match the search_params criterion

    def get_children(self):
        return self.children 

    def get_text(self):
        if self.still_searching and self.pipe:
            return self.text + ' ...'
        elif not self.still_searching and self.children == [ ]:
            return self.text + ': '+self.empty_text
        else:
            return self.text

    def get_comparator(self):
        return (-1e10, 0, string.lower(self.text)) 

    def update(self, app):
        any = 0

        if not self.pipe:
            self.pipe = self.pipe_getter()
            any = 1

        if self.pipe.finished() and self.still_searching:
            self.still_searching = 0
            return 1

        list = self.pipe.read_all()

        prev_pair = None
        for pair in list:
            if pair == prev_pair:
                continue
            link, item = prev_pair = pair

            # this is bad, really
            try:
                item = utility.check_and_demangle_item(item)
            except:
                self.empty_text = item
                continue

            keywords = item.get('keywords',[ ])

            if type(keywords) != types.ListType:
                continue

            # Comparing a string to a unicode can throw an error,
            # so force everything to unicode
            for i in range(len(keywords)):
                keywords[i] = utility.force_unicode(keywords[i])

            bad = 0                
            for key in self.keywords:
                if key not in keywords:
                    bad = 1
            for key in self.anti_keywords:
                if key in keywords:
                    bad = 1
            if item.get('type')=='file':
                if self.mime:
                    mime = item.get('mime')
                    if not mime:
                        bad=1
                    else:
                        if mime.find(self.mime) == -1:
                            bad = 1

            if item.get('type') != 'identity' and self.online_only:
                bad = 1
                            
            if item.get('type') == 'file' and self.remote_only :
                if app.node.data.has_key(item.get('name')+app.node.salt):
                    bad=1
            
            if bad:
                continue

            any = 1

            if item.get('type') == 'directory':
                any=1
                def pipe_getter(path=item['path'],app=app,link=link):
                    pipe = app.node.search_browse_task(link,path)
                    return pipe
                
                new_interior = Search_tree_directory(
                    pipe_getter,self.keywords,self.anti_keywords,\
                    self.what,self.mime,item['filename'],\
                    self.empty_text,self.single_text,0,self.show_url,self.show_number)
                new_interior.item=item
                
                if self.nickname:
                    new_interior.nickname=self.nickname
                    d=self.nickname+':'
                    if item['path']:
                        d=d+utility.force_unicode(string.join(item['path'],'/'))+'/'
                    visited_directories.append(d)

                self.children.append( new_interior )
            else:
                for child in self.children:
                    if child.matches(item,link):
                        child.add_source(item,link)
                        break
                else:
                    try:
                        if search_item_table.has_key(item['type']):
                            self.children.append( search_item_table[item['type']](item,link,self) )
                    except:                     
                        pass
                    
        return any

    def depth(self):
        return 0

    def show_expanded(self, chat, first_field):
        """Note: there is no point using generators here: gui stuff
        there is a bug in the ordering, due to calls to chat.show_before, show_after.
        """

        def show_expanded_thread(self, chat, first_field):
            import time

            #this dummy_action is so that the fields do not get deleted
            def dummy_action(d1,d2,d3):
                return
            self.first_field=first_field
            self.first_field.set_action(dummy_action)        
            self.last_field=first_field.child()
            self.last_field.set_action(dummy_action)
            size=0
            stats=''
            self.still_searching=1
            sleep_time= 0.01
            for item in self.children:
                item.in_list=0
            while self.list:
                self.list.remove(self.list[0])
            while self.still_searching:

                yield 'sleep',sleep_time
                
                sleep_time=sleep_time*1.25
                self.update(chat.app)
                children = self.get_children()
                if children and not self.show_number:
                    children = children[:]
                    children.sort(lambda x,y: cmp(x.get_comparator(),y.get_comparator()))
                field=self.first_field

                for item in children:
                    if not hasattr(item,'in_list'):
                        item.in_list=0                    
                    if item.in_list:
                        field=item.field
                    else:

                        if item.item['type']=='directory':
                            item.expanded=0
                        elif item.item['type']== 'file':
                            size=size+item.item['length']

                        field=field.child()
                        if not field:
                            return
                        item.field=field
                        item.field.set_action(dummy_action)
                        # ok, this is ugly...
                        if item.item['type']=='file' and item.item.get('name'):
                            for d in chat.app.file_server.downloaders:
                                if d.data['name']==item.item['name']:
                                    d.fields.append(item.field)

                        self.list.append(item)
                        if self.show_number:
                            number=len(self.list)
                        else:
                            number=0
                        item.show(self.depth(),chat,self.show_url,number)
                        item.in_list=1
                    stats='    '*self.depth()+'%d %s'%(len(self.list),self.what)
                    if self.what=='files':
                        stats=stats +', '+utility.human_size(size)
                if not self.list:
                    stats='    '*self.depth()+self.empty_text
                if len(self.list)==1:
                    stats='    '*self.depth()+self.single_text                
                self.last_field.show(stats+'...\n','grey')
                self.update(chat.app)

            self.last_field.show(stats+'.\n','grey')
            self.first_field.close()
            #now close invisible fields
            for item in self.list:
                item.field.close()
                #item.address_field.close()
            self.last_field.close()
            self.pipe.stop()

        utility.start_thread(show_expanded_thread(self,chat,first_field))

    def stop(self):
        Search_tree_object.stop(self)
        if self.pipe:
            self.pipe.stop()


class Search_tree_directory(Search_tree_interior):

    def depth(self):
        return len(self.item['path'])

    def show(self, depth, chat, show_url, number):
        def on_click(field,widget,event,root=self):
            return chat.directory_field_on_click(field,widget,event,root)
        chat.show_before(self.field, '    '*depth)
        if number:
            chat.show_before(self.field, '%d. '% number)

        chat.set_tabs(len(self.get_text())+3)            
        chat.show_before(self.field, utility.force_unicode(self.get_text()),'files',on_click)
        self.file_field = self.field.prev
        self.file_field.leaf = self
        chat.show_before(self.field,'\t--')
        chat.show_before(self.field,'\tdirectory\n')


    def get_actions(self,app):

        def download(b,_self=self,_app=app):
            _self.download(_app)
            app.download_manager.assert_visible()
            
        button1 = gtk.Button(_("Download directory"))
        button1.connect("clicked",download)
        buttons = [button1]
        return buttons

    def download(self, app, public=1, play=0):
        """
        If play == 1, download only the music files of the directory,
        and play them. The idea is to play a whole album.
        """

        def download_thread(self, app, public, play):
            #this is gui too... should not use generators

            dirname = app.daemon.config['download_dir']
            
            for i in self.item['path']:
                dirname = os.path.join(dirname, i)
                if not os.path.exists(dirname):
                    os.mkdir(dirname)

            list = []
            self.still_searching=1
            sleep_time=0.1
            if play:
                app.music_manager.clear()

            while self.still_searching:
                yield 'sleep',sleep_time
                sleep_time=sleep_time*1.25
                children = self.get_children()
                for item in children:
                    if item not in list:
                        list.append(item)                    
                        if item.item['type']=='file':
                            if not play:
                                app.file_server.download(item,dirname,no_overload=1)
                            else:
                                item.set_player(app)
                                if item.player.players.has_key('Add to playlist'):
                                    item.play('Add to playlist',app)
                                    app.music_manager.play_music()
                        else:
                            item.download(app)

                self.update(app)
                
        utility.start_thread(download_thread(self, app, public, play))


class Search_tree_leaf(Search_tree_object):
    
    def __init__(self, item, source, father = None):
        Search_tree_object.__init__(self)

        self.father  = father
        self.item    = item
        self.sources = [ source ]
        self.widget  = None
        self.field = None
        self._setup()

    def get_text(self):
        return self.title

    def get_comparator(self):
        return (-len(self.sources), 0, string.lower(self.title))

    def get_description(self, node_address):
        return self.description

    def get_widget(self):
        return self.widget

    def add_source(self,item,address):
        if address in self.sources:
            return 0
        self.sources.append(address)
        return 1
 
    def browse(self, app):
        search_browse_files(self.sources[0], app.node,app)

class Search_tree_identity(Search_tree_leaf):
    def __init__(self, item,source,father):
        self.address = None
        self.address_field = None
        Search_tree_leaf.__init__(self, item,source,father)

    def _setup(self):
        if self.item['type'] == 'identity':
            self.address = self.sources[0]

        self.name = utility.force_unicode(self.item['name']) + \
                    ' (' + utility.force_unicode(self.item.get('human-name','')) + ')'

        self.update_description()


    def get_widget(self):
        if self.widget is None:
            hash = name_server.key_name(self.item['key'])
            self.widget = widgets.Signature(hash)
        # Proof of @R43: hash assigned from name_server.key_name; @E25; and not shared
        # with any other thread or written to.
        return self.widget

    def update_description(self):
        self.title = self.name
        self.description = _('Name: ') + \
                           utility.force_unicode(self.item.get('human-name','')) + \
                           '\n  '+_('Username: ') + \
                           utility.force_unicode(self.item['name'])
                                             
        if self.address:
            self.title = self.title + _(' (online)')
            self.description = self.description + \
                '\n  '+_('Location: ')+ self.address[0]+':'+repr(self.address[1]) 

    def get_comparator(self):
        return (0, 0, string.lower(self.title))

    def add_source(self,item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        if item['type'] == 'identity':
            self.address = address
            self.update_description()
            if self.address_field:
                self.show_address()
        return 1

    def get_actions(self,app):
        if app.name_server.acquaintances.has_key(name_server.key_name(self.item['key'])):
            str = _("View")
        else:
            str = _("Add to contact list")
        button1 = gtk.Button(str)
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _app.name_server.edit_acquaintance(_self.item,_self.address))

        if not self.address:
            return [ button1 ]

        button2 = gtk.Button(_("Browse their files"))
        button2.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.browse(_app))

        return [ button1, button2 ]

    def matches(self, item,address):
        return item['type'] in ['identity','identity offline demangled'] and self.item['key'] == item['key']

    #todo: have different classes of chat fields for names, addresses etc...
    def show_address(self):
        if self.address:
            self.address_field.set_action(self.chat.host_field_on_click)
            color = 'links'            
            if self.show_ip:
                host = self.address[0]
            else:
                import socket
                try:
                    host = socket.gethostbyaddr(self.address[0])[0]
                except:
                    host = self.address[0]

            text = '%s:%d'%(host,self.address[1])
        else:
            text='offline'
            color='grey'
        self.address_field.show(text,color)

    def show(self, depth, chat, show_url, number):
        """self.field should not be closed now"""
        chat.show_before(self.field, '    '*depth)
        if number:
            chat.show_before(self.field, '%d. '% number)
        field = chat.get_field_before(self.field)
        field.data = self.item
        field.acq = None
        field.set_action(chat.name_field_on_click)
        field.show('%s'  % self.item.get('name'),'people')
        field.close()
        
        if self.item.get('human-name'):
            chat.show_before(self.field, ' (%s)'% self.item.get('human-name'))

        if show_url:
            if show_url==1:
                self.show_ip=1
            else:
                self.show_ip=0
            chat.show_before(self.field, '\t ')
            self.address_field = chat.get_field_before(self.field)
            self.chat=chat
            self.show_address()

        chat.show_before(self.field, '\n')

        chat.set_tabs(len(self.item.get('name')+self.item.get('human-name'))+6)


class Search_tree_daemonic_node(Search_tree_leaf):
    def _setup(self):
        self.title = _('Daemon at %s')%self.item['name']
        self.description = self.title

    def get_comparator(self):
        return (-len(self.sources), 1, string.lower(self.title))
    
    def get_actions(self,app):
        button1 = gtk.Button(_("Browse their files"))
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.browse(_app))

        return [ button1 ]
 
class Search_tree_channel_exists(Search_tree_leaf):
    def _setup(self):
        self.title = utility.force_unicode( self.item['name'] )
        self.description = _("1 subscriber on-line.")

    def add_source(self, item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        self.description = _("%d subscribers on-line.") % len(self.sources)
        return 1

    def get_actions(self,app):
        if app.channels.list.has_key(self.item['name']):
            button1 = gtk.Button(_("Unsubscribe"))
            button1.connect("clicked",\
                lambda _b, _self=self,_app=app:
                    _app.channels.unsub(_self.item['name']))
        else:
            button1 = gtk.Button(_("Subscribe"))
            button1.connect("clicked",\
                lambda _b, _self=self,_app=app:
                    _app.channels.sub(_self.item['name']))

        button2 = gtk.Button(_("List subscribers"))
        button2.connect("clicked",\
            lambda _b, _self=self,_app=app: search_for_people_on_channel(_self.item['name'],_app.node,_app))

        return [ button1, button2 ]
 
    def matches(self, item,address):
        return self.item['type'] == 'channel exists' and self.item['name'] == item['name']


class Search_tree_file(Search_tree_leaf):

    def _setup(self):
        if not self.item.has_key('filename') or \
             not self.item.has_key('length'):
             #or not self.item.has_key('name'):
            raise error.Error(_('Bad datum'))

        #if self.item.get('music_title'):
        #    self.title = self.item['music_title']
        #    if self.item.get('music_artist'):
        #        self.title = self.title + _(' by ') + self.item['music_artist']
        #else:
        #    self.title = self.item['filename']

        self.title = utility.force_unicode(self.item['filename'])

        self.file_field = None

        #self.is_music = self.player.playable
        self.title = utility.force_unicode(self.title)         
        keys = [('filename',''),('music_title',_('   Title: ')),('music_album',_('   Album: ')),('music_artist',_('   Artist: '))]
        self.description = ''
        for key in keys:
            if self.item.has_key(key[0]):
                self.description = self.description+key[1]+\
                    utility.force_unicode(self.item[key[0]])+'\n'
        self.description_template = self.description
        if self.sources[0][0] == 'overnet':
            self.description = self.description + _('File available from Overnet.')



    def get_description(self, node_address):
        if self.sources[0][0] == 'overnet':
            return self.description

        if node_address in self.sources:
            n = max(1,len(self.sources)-1)
        else:
            n = len(self.sources)

        return self.description_template + \
            (_("File available from %d source%s.") % (n,(n>1 and 's' or '')))

    def add_source(self, item,address):
        if not Search_tree_leaf.add_source(self,item,address):
            return 0

        return 1
 
    def set_player(self,app):
        if settings.gtk_loaded():
            self.player = Player(self.item['filename'],app)

    def get_widget(self):
        if self.widget is None:
            if self.sources[0][0] != 'overnet':
                self.widget = gtk.HBox(gtk.FALSE, 0)
                self.widget.pack_start(gtk.Label(_('URL: ')), gtk.FALSE, gtk.FALSE, 0)
                entry = gtk.Entry()
                entry.set_text(hash.hash_to_url(self.item['name']))
                self.widget.pack_start(entry, gtk.TRUE, gtk.TRUE, 0)
                self.widget.show_all()
        return self.widget


    def get_actions(self,app):
        
        button1 = gtk.Button(_("Download"))
        button1.connect("clicked",\
            lambda _b, _self=self,_app=app: _self.download(_app))
        buttons = [button1]
        if self.sources[0][0] != 'overnet':
            button2 = gtk.Button(_("View sources"))
            button2.connect("clicked",\
               lambda _b, _self=self,_app=app: _self.view_sources(_app))
            buttons.insert(0,button2)
            self.set_player(app)
            for item in self.player.actions():
                button = gtk.Button(item)
                button.connect("clicked", lambda b, self=self,app=app,item=item: self.play(item,app))
                buttons.insert(0,button)
        return buttons

    def matches(self, item,address):

        if self.item.get('name'):            
            return self.item.get('name') == item.get('name')        
        else:            
            return self.item == item        
        

    def download(self, app):
        try:
            app.file_server.download(self,app.daemon.config['download_dir'])
            app.download_manager.assert_visible()
        except error.Error, err:
            app.show_error(err)

    def view_sources(self, app):
        search_show_sources(self.sources, app.node,app)

    def get_text(self):
        return self.title + '  (%s) '  % utility.human_size(self.item['length'])

    def show(self, depth, chat, show_url,number):
        """show the file properties
        self.field is not closed here, but in the show_expanded method """

        self.set_player(chat.app)
        chat.show_before(self.field, '    '*depth)

        if number:
            str='%d. '%number
        else:
            str=''
            
        chat.show_before(self.field, str)
        chat.show_before(self.field, utility.force_unicode(self.title),'files',chat.file_field_on_click)
        self.file_field = self.field.prev
        self.file_field.leaf = self
        chat.show_before(self.field, ' \t')
        chat.set_tabs(len(self.title)+3)            
        chat.show_before(self.field, utility.human_size(self.item['length'])+' \t')

        if self.item.get('mime'):
            mime_str=self.item.get('mime').replace(' \x08','')
        else:
            mime_str='unknown    '
        if self.sources[0][0]=='overnet':
            mime_str=mime_str+', overnet'
        chat.show_before(self.field,mime_str+' \t','grey')
        if self.item.get('bitrate'):
            chat.show_before(self.field,' %dKb/s'%self.item.get('bitrate')+' \t','grey')

        #chat.show_before(self.field,'\n')
        self.field.show('\n')


    def play(self, action, app):
        """
        If self.item does not have a 'name', it must have a url.
        it means it was returned by a 'files availables' query.
        in that case I should know where it is, and be able to
        download it without querying the hashtable.
        see /give
        """

        #first check if the file is local

        #first case: local file accessed through /ls
        if self.item.get('local_path'):
            player = self.player.play(action, self.item, None,self.file_field)
            return

        #2nd case: locally published file
        if app.file_server.paths.has_key(self.item['name']):
            path, mtime = app.file_server.paths.get(self.item['name'])
            try:
                self.item['local_path'] = path
                player = self.player.play(action, self.item, None, self.file_field)
            except error.Error, err:
                app.show_error(err)
                # and don't show checkbox...            
            return

        # file is not local. we will have to download it
        
        #first check if already downloading
        for d in app.file_server.downloaders:
            if d.data['name']==self.item['name']:
                downloader=d
                break
        else:
            try:
                downloader = app.file_server.download(self,app.daemon.config['download_dir'])
            except error.Error, err:
                app.show_error(err)
                return

        try:
            player = self.player.play(action, self.item, downloader, self.file_field)
        except error.Error, err:
            app.show_error(err)
            if downloader.running:
                downloader.stop()
            app.file_server.remove_file(downloader.data['name'],downloader.filename)
            return
        
        # on player close: unlink file (maybe)
        def on_close(window,player=player,downloader=downloader,filename=downloader.filename,app=app):
            # Note: this only works because downloader stops in the gtk thread
            if not player.keep_file_button.get_active():
                if downloader.running:
                    downloader.stop()
                if downloader.success:
                    app.file_server.remove_file(downloader.data['name'],downloader.filename)

        #music_manager handles destroy events by himself...
        #if player != app.music_manager.window:
        #    player.connect("destroy",on_close)


search_item_table = {
    'identity' : Search_tree_identity,
    'identity offline demangled' : Search_tree_identity,
    'daemonic node' : Search_tree_daemonic_node,
    'channel exists' : Search_tree_channel_exists,
    'file' : Search_tree_file
}

class Searcher:
    def __init__(self, title, pipe, node, app, keywords=[], anti_keywords=[]):
        self.title = title
        self.pipe  = pipe
        self.keywords  = keywords
        self.anti_keywords = anti_keywords
        
        self.node  = node
        self.app   = app
        self.current_selection = None
        self.actions = [ ]

        self.running = 0

    def start(self):
        self.running = 1

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL) 
        self.window.set_default_size(400,400)
        #self.window.set_border_width(10)        
        self.window.set_title(self.title)
        self.window.connect("destroy",\
            lambda _b, _self=self: _self.stop())

        vbox = gtk.VBox(gtk.FALSE, 5)
        self.window.add(vbox)

        #self.label = gtk.Label('Searching...')
        #self.label.set_alignment(0,0)
        #vbox.pack_start(self.label, gtk.FALSE, gtk.FALSE, 0)

        def on_selection_change(new_selection, self=self):
            self.bin.foreach(lambda c, self=self: self.bin.remove(c))

            for item in self.actions:
                self.hbox.remove(item)
            self.actions = [ ]

            if not new_selection:
                self.description_label.set_text('')
                self.vbox_2.queue_resize()
                return

            self.actions = new_selection[-1].get_actions(self.app)
            for item in self.actions:
                item.set_flags(gtk.CAN_DEFAULT)
                self.hbox.pack_start(item, gtk.TRUE, gtk.TRUE, 0)
            self.hbox.show_all()
            
            if self.actions:
                self.actions[0].grab_default()

            self.description_label.set_text(new_selection[-1].get_description(self.node.address))
            widget = new_selection[-1].get_widget()
            if widget:
                self.bin.add(widget)
                widget.show()

            self.vbox_2.queue_resize()

        def child_getter(path):
            children = path[-1].get_children()
            if not children:
                return children

            children = children[:]
            children.sort(lambda x,y: cmp(x.get_comparator(),y.get_comparator()))
            return children

        self.root = Search_tree_interior(lambda self=self: self.pipe, self.keywords, self.anti_keywords,
                                         'files', '', self.title, _('couldn\'t find anything, sorry'))
        self.tree_widget = widgets.Tree([self.root],child_getter,lambda x: x.get_text(),on_selection_change)

        vbox.pack_start(self.tree_widget.packee(), gtk.TRUE, gtk.TRUE, 0)

        #def on_select(_b, row,column,event, _self=self): 
        #  _self.current_selection = row
        #  str = _self.results[row].description
        #  _self.description_label.set_text(str)
# 
#      _self.bin.foreach(lambda c, _self=_self: _self.bin.remove(c))
#      if _self.results[row].widget:
#        _self.bin.add(_self.results[row].widget)
#        _self.results[row].widget.show()
#
#      for item in _self.actions:
#        _self.hbox.remove(item)
#      _self.actions = _self.results[row].get_action_bar(_self.app)
#      for item in _self.actions:
#        _self.hbox.pack_start(item, gtk.TRUE, gtk.TRUE, 0)
#      _self.hbox.show_all()
#      
#      if event and event.type == gtk.gtk._2BUTTON_PRESS:
#        if len(_self.actions):
#          _self.actions[0].emit("clicked",None)
#
#    scrolly = gtk.ScrolledWindow()
#    scrolly.set_policy(gtk.POLICY_NEVER,gtk.POLICY_AUTOMATIC)
#    vbox.pack_start(scrolly)
#    
#    self.list_box = gtk.CList(1)
#    self.list_box.set_selection_mode(gtk.SELECTION_BROWSE)
#    self.list_box.connect("select-row", on_select)
#    #self.list_box.connect("unselect-row", on_unselect)
#    scrolly.add(self.list_box)

        self.vbox_2 = gtk.VBox(gtk.FALSE, 5)
        self.vbox_2.set_border_width(10)
        vbox.pack_start(self.vbox_2, gtk.FALSE, gtk.FALSE, 0)
        
        self.description_label = gtk.Label('')
        self.description_label.set_justify(gtk.JUSTIFY_LEFT)
        self.description_label.set_alignment(0,0)
        self.vbox_2.pack_start(self.description_label, gtk.FALSE, gtk.FALSE, 0)

        self.bin = gtk.Alignment(1,1,0.5,1)
        self.vbox_2.pack_start(self.bin, gtk.FALSE, gtk.FALSE, 0)

        # Obscure GTK sizing bug workaround
        self.alignment = gtk.Alignment(0,0,1,1)
        self.vbox_2.pack_end(self.alignment, gtk.FALSE, gtk.FALSE, 0)
        self.hbox = gtk.HBox(gtk.FALSE, 5)
        self.alignment.add(self.hbox)

        self.close_button = gtk.Button(_("Close"))
        self.close_button.connect("clicked",\
            lambda _b, _window=self.window: _window.destroy())
        self.close_button.set_flags(gtk.CAN_DEFAULT)
        self.hbox.pack_end(self.close_button, gtk.TRUE, gtk.TRUE, 0)
        
        #button = gtk.Button("View sources")
        #button.connect("clicked",\
        #  lambda _b, _self=self: _self.view_sources())
        #hbox.pack_end(button, gtk.TRUE, gtk.TRUE, 0)

        #button = gtk.Button("Download")
        #button.connect("clicked",\
        #  lambda _b, _self=self: _self.download())
        #hbox.pack_end(button, gtk.TRUE, gtk.TRUE, 0)

        self.app.show_window(self.window, 'search')
    
        self.update(250.0)
        #utility.schedule_mainthread(250.0, self.update)
        #timeout_add(250,self.update)

    def stop(self):

        list = [ self.root ]
        while list:
            children = list[0].get_children()
            if children:
                list.extend(children)
            list[0].stop()
            del list[0]

        self.running = 0
 
    def update(self, last_time):
        assert(type(last_time) == types.FloatType)
        if not self.running:
            return

        any = 0
        for item in self.tree_widget.current_path:
            any = item.update(self.app) or any

        if any:
            self.tree_widget.refresh_display()

        new_time = last_time * 5.0 / 4.0
        if new_time > 10000.0:
            new_time = 10000.0

        gtk.timeout_add( int(new_time), self.update, new_time)
        
        #return self.running

    #  for item in self.pipe.read_all():
    #    self.insert(item[0],item[1])

    #  finished = self.pipe.finished()

    #  if finished:
    #    self.label.set_text('Search results:')

    #  return not finished

    #def insert(self, link,item):
    #  if type(item) != types.DictType or \
    #     not item.has_key('type') or \
    #     not item.has_key('keywords') or \
    #     type(item['keywords']) != types.ListType:
    #    return
#
#    for key in self.list:
#      if key not in item['keywords']:
#        return
#    
#    for result in self.results:
#      if result.matches(item,link):
#        result.add_source(link)
#       return
#    
#    new_result = Search_result(item,link)
#  
#    self.results.append(new_result)
#    self.list_box.append([new_result.title])


# vim: set expandtab :
